# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015, 2016, 2020 Didier Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import operator
import re
from collections import OrderedDict
from functools import reduce

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.messages import warning as warning_message
from django.contrib.messages.views import SuccessMessageMixin
from django.contrib.sites.models import Site
from django.core.exceptions import PermissionDenied
from django.db.models import Case, Count, F, IntegerField, Q, When
from django.http import HttpResponseRedirect
from django.template.defaultfilters import date, time
from django.template.loader import render_to_string
from django.urls import Resolver404, reverse, reverse_lazy
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.text import slugify
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy as _p
from django.views.generic import ListView, RedirectView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView

from django_ical.views import ICalFeed
from rolepermissions.checkers import has_role
from rolepermissions.mixins import HasPermissionsMixin
from tablib import Dataset

from apps.common import (
    DV_SEASON_AUTUMN,
    DV_SEASON_CHOICES,
    DV_SEASON_LAST_SPRING_MONTH,
    DV_SEASON_SPRING,
    DV_SEASON_STATE_OPEN,
    DV_SEASON_STATE_RUNNING,
    DV_SEASON_STATES,
    DV_STATES,
)
from apps.common.views import ExportMixin
from apps.user import FORMATION_KEYS, FORMATION_M1, FORMATION_M2
from apps.user.models import USERSTATUS_ACTIVE, USERSTATUS_DELETED, USERSTATUS_RESERVE
from apps.user.views import ActorsList, HelpersList
from defivelo.roles import has_permission, user_cantons
from defivelo.templatetags.dv_filters import dv_season_url
from defivelo.views import MenuView

from .. import (
    AVAILABILITY_FIELDKEY,
    CHOICE_FIELDKEY,
    CHOSEN_AS_ACTOR,
    CHOSEN_AS_HELPER,
    CHOSEN_AS_LEADER,
    CHOSEN_AS_NOT,
    CHOSEN_KEYS,
    CONFLICT_FIELDKEY,
    SEASON_WORKWISH_FIELDKEY,
    STAFF_FIELDKEY,
    SUPERLEADER_FIELDKEY,
)
from ..forms import (
    SeasonAvailabilityForm,
    SeasonAvailabilityReminderForm,
    SeasonForm,
    SeasonNewHelperAvailabilityForm,
    SeasonStaffChoiceForm,
    SeasonToSpecificStateForm,
)
from ..forms.season import SeasonStaffFilterForm
from ..models import HelperSessionAvailability, Qualification, Season
from ..models.availability import HelperSeasonWorkWish
from ..models.qualification import (
    CATEGORY_CHOICE_A,
    CATEGORY_CHOICE_B,
    CATEGORY_CHOICE_C,
)
from ..utils import (
    GeneralSeason,
    get_users_roles_for_session,
    seasons_in_scope_for_user,
)
from .mixins import CantonSeasonFormMixin

EXPORT_NAMETEL = gettext("{name} - {tel}")


class SeasonMixin(CantonSeasonFormMixin, MenuView):
    required_permission = "challenge_season_crud"
    model = Season
    context_object_name = "season"
    form_class = SeasonForm
    raise_without_cantons = True

    def get_context_data(self, **kwargs):
        context = super(SeasonMixin, self).get_context_data(**kwargs)
        # Add our menu_category context
        context["menu_category"] = "season"
        context["season"] = self.season
        return context

    def get_form_kwargs(self):
        form_kwargs = super(SeasonMixin, self).get_form_kwargs()
        if has_permission(self.request.user, "cantons_all"):
            form_kwargs["cantons"] = DV_STATES
        else:
            # Ne permet que l’édition et la création de moiss pour les cantons gérés
            form_kwargs["cantons"] = self.request.user.managedstates.all().values_list(
                "canton", flat=True
            )
        return form_kwargs

    def get_queryset(self):
        if self.model == Season:
            return self.request.user.profile.get_seasons(
                self.raise_without_cantons
            ).prefetch_related("leader")

        qs = super(SeasonMixin, self).get_queryset()

        if self.model == get_user_model():
            try:
                usercantons = user_cantons(self.request.user)
            except LookupError:
                if self.raise_without_cantons:
                    raise PermissionDenied
                return qs
            # Check that the intersection isn't empty
            cantons = list(set(usercantons).intersection(set(self.season.cantons)))
            if not cantons:
                raise PermissionDenied
        return qs


class SeasonListRedirect(RedirectView):
    is_permanent = False

    def get_redirect_url(self, *args, **kwargs):
        return dv_season_url()


class SeasonListView(SeasonMixin, ListView):
    allow_empty = True
    allow_future = True
    context_object_name = "seasons"
    make_object_list = True
    raise_without_cantons = False

    def dispatch(self, request, *args, **kwargs):
        self.year = int(self.kwargs.pop("year"))
        self.dv_season = self.kwargs.pop("dv_season", None)
        if self.dv_season:
            self.dv_season = int(self.dv_season)
            if self.dv_season not in [s[0] for s in DV_SEASON_CHOICES]:
                raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        qs = super().get_queryset().filter(year=self.year)
        if self.dv_season == DV_SEASON_SPRING:
            qs = qs.filter(month_start__lte=DV_SEASON_LAST_SPRING_MONTH)
        elif self.dv_season == DV_SEASON_AUTUMN:
            qs = qs.filter(month_start__gt=DV_SEASON_LAST_SPRING_MONTH)
        return qs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["year"] = self.year
        if self.dv_season:
            context["dv_season"] = self.dv_season
            if self.dv_season == DV_SEASON_SPRING:
                context["dv_season_prev_day"] = datetime.date(
                    self.year - 1, DV_SEASON_LAST_SPRING_MONTH + 1, 1
                )
                context["dv_season_next_day"] = datetime.date(
                    self.year, DV_SEASON_LAST_SPRING_MONTH + 1, 1
                )
            elif self.dv_season == DV_SEASON_AUTUMN:
                context["dv_season_prev_day"] = datetime.date(self.year, 1, 1)
                context["dv_season_next_day"] = datetime.date(self.year + 1, 1, 1)
            else:
                # Will never happen
                raise PermissionDenied
        return context


class SeasonDetailView(SeasonMixin, DetailView):
    allow_season_fetch = True
    raise_without_cantons = False

    def dispatch(self, request, *args, **kwargs):
        """
        Allow view by participants of the season if allowed
        """
        try:
            allowed = False
            if self.season.unprivileged_user_can_see(request.user):
                # Season participants
                allowed = True
            elif has_permission(request.user, self.required_permission):
                # State Managers
                allowed = True

            if allowed:
                return super().dispatch(request, *args, **kwargs)
        except ValueError:
            raise Resolver404
        raise PermissionDenied

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Add our submenu_category context
        context["submenu_category"] = "season-detail"
        sessions = self.season.sessions_by_orga
        if not has_permission(self.request.user, "challenge_see_all_orga"):
            sessions = sessions.filter(orga__coordinator=self.request.user)
        context["sessions_by_orga"] = sessions
        return context


class SeasonUpdateView(SeasonMixin, SuccessMessageMixin, UpdateView):
    success_message = _("Mois mis à jour")

    def dispatch(self, request, bypassperm=False, *args, **kwargs):
        if (
            bypassperm
            or
            # Soit j'ai le droit
            has_permission(request.user, self.required_permission)
        ):
            return super(SeasonUpdateView, self).dispatch(request, *args, **kwargs)
        else:
            raise PermissionDenied


class SeasonHelpersMixin(SeasonMixin):
    """
    Provide helper functions to list helpers for season in various formats
    """

    def potential_helpers_qs(self, qs=None):
        """
        Queryset of potential helpers
        """
        if not qs:
            qs = get_user_model().objects.exclude(profile__status=USERSTATUS_DELETED)
            if self.season:
                seasoncantons = self.season.cantons
                # S'il y au moins un canton en commun
                cantons_filter = [
                    Q(profile__activity_cantons__overlap=seasoncantons)
                ] + [Q(profile__affiliation_canton__in=seasoncantons)]
                qs = qs.filter(reduce(operator.or_, cantons_filter))

            # Pick the one helper from the command line if it makes sense
            resolvermatch = self.request.resolver_match
            try:
                qs = qs.filter(pk=int(resolvermatch.kwargs["helperpk"]))
            except (KeyError, TypeError):
                pass
        return qs.prefetch_related(
            "profile",
            "profile__actor_for",
            "profile__actor_for__translations",
        )

    def potential_helpers(self, qs=None):
        """
        Return the struct of potential helpers, in three blocks
        """
        qs = self.potential_helpers_qs(qs)
        all_helpers = qs.order_by("first_name", "last_name")
        return (
            (
                _("Moniteur·trice·s 2"),
                all_helpers.filter(profile__formation=FORMATION_M2),
            ),
            (
                _("Moniteur·trice·s 1"),
                all_helpers.filter(profile__formation=FORMATION_M1),
            ),
            (
                _("Intervenant·e·s"),
                all_helpers.filter(profile__formation="").exclude(
                    profile__actor_for__isnull=True,
                ),
            ),
        )

    def current_availabilities(self):
        if not hasattr(self, "object"):
            self.object = self.get_object()
        return HelperSessionAvailability.objects.filter(
            session__in=self.object.sessions_with_qualifs
        ).prefetch_related("helper")

    def current_availabilities_present(self):
        return self.current_availabilities().exclude(availability="n")

    @cached_property
    def available_helpers(self):
        # Only take available people
        # Fill in the helpers with the ones we currently have
        helpers_pks = self.current_availabilities_present().values_list(
            "helper_id", flat=True
        )
        return self.potential_helpers(
            qs=get_user_model().objects.filter(pk__in=helpers_pks)
        )

    @cached_property
    def season_helpers(self):
        """
        All the helpers who filled availabilities for this season
        """
        helpers_pks = self.current_availabilities_present().values_list(
            "helper", flat=True
        )
        return get_user_model().objects.filter(pk__in=helpers_pks)


class SeasonAvailabilityMixin(SeasonHelpersMixin):
    view_is_update = False
    view_is_planning = False

    def dispatch(self, request, *args, **kwargs):
        if self._has_default_access(request) or self.custom_access_allowed(request):
            return super().dispatch(request, bypassperm=True, *args, **kwargs)
        raise PermissionDenied

    def _has_default_access(self, request):
        return (
            # Check that the request user is alone in the potential_helpers and we're in season_update
            (
                self.potential_helpers_qs()
                .filter(
                    Q(profile__formation__in=FORMATION_KEYS)
                    | Q(profile__actor_for__isnull=False)
                )
                .filter(pk=request.user.pk)
                .exists()
                and self.season
                and (
                    (
                        self.season.staff_can_update_availability
                        and not self.view_is_planning
                    )
                    or (self.season.staff_can_see_planning and self.view_is_planning)
                )
            )
            or
            # Soit j'ai le droit et c'est le bon moment
            (
                has_permission(request.user, self.required_permission)
                and (not self.view_is_update or self.season.manager_can_crud)
            )
        )

    def custom_access_allowed(self, request):
        # Overridden in aggregated/general-planning views
        return False

    def get_initial(self, all_hsas=None, all_helpers=None):
        initials = OrderedDict()
        if not all_hsas:
            all_hsas = self.current_availabilities()
        if not all_helpers:
            all_helpers = self.potential_helpers()

        if all_hsas:
            all_sessions = self.object.sessions_with_qualifs
            all_wishes = dict(
                self.object.work_wishes.values_list("helper_id", "amount")
            )
            potential_conflicts = (
                HelperSessionAvailability.objects
                # Ne boucle pas
                .exclude(session__id__in=all_sessions)
                # Seulement des sessions concernées
                .filter(session__day__in=all_sessions.values_list("day", flat=True))
                # Seulement dans les états qui nous intéressent
                .filter(chosen_as__in=CHOSEN_KEYS)
                .prefetch_related("session", "session__orga")
            )
            for helper_category, helpers in all_helpers:
                helpers_conflicts = list(potential_conflicts.filter(helper__in=helpers))

                for helper in helpers:
                    helper_availability = {
                        a.session_id: a for a in all_hsas if a.helper == helper
                    }
                    # Fill in the wishes
                    initials[SEASON_WORKWISH_FIELDKEY.format(hpk=helper.pk)] = (
                        all_wishes[helper.pk] if helper.pk in all_wishes else 0
                    )

                    for session in all_sessions:
                        fieldkey = AVAILABILITY_FIELDKEY.format(
                            hpk=helper.pk, spk=session.pk
                        )
                        staffkey = STAFF_FIELDKEY.format(hpk=helper.pk, spk=session.pk)
                        choicekey = CHOICE_FIELDKEY.format(
                            hpk=helper.pk, spk=session.pk
                        )
                        conflictkey = CONFLICT_FIELDKEY.format(
                            hpk=helper.pk, spk=session.pk
                        )
                        superleaderkey = SUPERLEADER_FIELDKEY.format(
                            hpk=helper.pk, spk=session.pk
                        )
                        try:
                            hsa = helper_availability[session.id]
                            initials[fieldkey] = hsa.availability
                            # Si un choix est fait _dans_ une session (qualif)
                            initials[staffkey] = session.user_assignment(helper)
                            initials[choicekey] = True
                            # Le choix n’est fait qu'au niveau de la session
                            if not initials[staffkey]:
                                initials[staffkey] = hsa.chosen_as
                                initials[choicekey] = False
                        except Exception:
                            initials[fieldkey] = ""
                            initials[staffkey] = ""
                            initials[choicekey] = ""
                        # List super-leaders (Moniteurs +)
                        if session.superleader == helper:
                            initials[superleaderkey] = True

                        # Trouve les disponibilités en conflit
                        initials[conflictkey] = [
                            hc.session
                            for hc in helpers_conflicts
                            if hc.helper_id == helper.pk
                            and hc.session.day == session.day
                        ]
            return initials

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        resolvermatch = self.request.resolver_match
        try:
            context["helperpk"] = resolvermatch.kwargs["helperpk"]
        except (KeyError, TypeError):
            pass

        context["user_can_see_season"] = self.season.unprivileged_user_can_see(
            self.request.user
        )
        # Add our submenu_category context
        context["submenu_category"] = "season-availability"
        context["sessions"] = self.object.sessions_with_qualifs.annotate(
            n_qualifs=Count("qualifications", distinct=True),
            n_leaders=Count(
                Case(
                    When(
                        availability_statuses__chosen_as=CHOSEN_AS_LEADER,
                        then=F("availability_statuses__id"),
                    ),
                    output_field=IntegerField(),
                ),
                distinct=True,
            ),
            n_helpers=Count(
                Case(
                    When(
                        availability_statuses__chosen_as=CHOSEN_AS_HELPER,
                        then=F("availability_statuses__id"),
                    ),
                    output_field=IntegerField(),
                ),
                distinct=True,
            ),
            n_actors=Count(
                Case(
                    When(
                        availability_statuses__chosen_as=CHOSEN_AS_ACTOR,
                        then=F("availability_statuses__id"),
                    ),
                    output_field=IntegerField(),
                ),
                distinct=True,
            ),
        )
        return context


class SeasonToStateMixin(SeasonHelpersMixin, SeasonUpdateView):
    form_class = SeasonToSpecificStateForm
    season_to_state = None
    template_name = "challenge/season_form_tostate.html"

    def get_form_kwargs(self):
        """
        Hand over tostate to the Form class
        """
        form_kwargs = super().get_form_kwargs()
        form_kwargs["tostate"] = self.season_to_state
        return form_kwargs

    def get_context_data(self, **kwargs):
        """
        Set tostate to the verbose desired state
        """
        context = super().get_context_data(**kwargs)
        context["tostate"] = next(
            (v for (k, v) in DV_SEASON_STATES if k == self.season_to_state)
        )
        context["recipients"] = self.season_helpers
        return context

    def get_initial(self):
        """
        Set the inital form value to the state we target
        """
        return {"state": self.season_to_state}

    def dispatch(self, request, bypassperm=False, *args, **kwargs):
        """
        Push away if we're already in the given state
        """
        if self.season and self.season.state == self.season_to_state:
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)


class SeasonToRunningView(SeasonToStateMixin):
    season_to_state = DV_SEASON_STATE_RUNNING

    def get_email(self, profile=None):
        """
        Get a simple struct with the email we're sending at this step
        """
        if profile:
            helperpk = profile.pk
        else:
            # Fake a profile that can be used in template
            profile = {"get_full_name": _("{Prénom} {Nom}")}
            helperpk = 0

        planning_link = self.request.build_absolute_uri(
            reverse(
                "season-planning",
                kwargs={"pk": self.season.pk, "helperpk": helperpk},
            )
        )

        return {
            "subject": settings.EMAIL_SUBJECT_PREFIX
            + gettext("Planning {season}").format(season=self.season.desc()),
            "body": {
                "pre": render_to_string(
                    "challenge/season_email_to_state_running_pre.txt",
                    {
                        "profile": profile,
                        "season": self.season,
                        "planning_link": planning_link,
                        "current_site": Site.objects.get_current(),
                    },
                ),
                "post": render_to_string(
                    "challenge/season_email_to_state_running_post.txt",
                    {
                        "profile": profile,
                        "season": self.season,
                        "planning_link": planning_link,
                        "current_site": Site.objects.get_current(),
                    },
                ),
            },
        }

    def dispatch(self, request, bypassperm=False, *args, **kwargs):
        """
        Push away if we're not allowed to go to that state now
        """
        if self.season and self.season.can_set_state_running:
            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["some_recipients_cant_login"] = any(
            not user.profile.can_login for user in context["recipients"]
        )
        context["email"] = self.get_email()
        return context

    def form_valid(self, form):
        """
        Run our specific action here
        """
        # Save first
        form_result = super().form_valid(form)
        if form.cleaned_data["sendemail"]:
            for helper in self.season_helpers:
                email = self.get_email(helper)
                body = "\n".join(
                    [
                        email["body"]["pre"],
                        form.cleaned_data["customtext"],
                        email["body"]["post"],
                    ]
                )
                helper.profile.send_mail(email["subject"], body)
        return form_result


class SeasonToOpenView(SeasonToStateMixin):
    season_to_state = DV_SEASON_STATE_OPEN

    def get_email(self, profile=None):
        """
        Get a simple struct with the email we're sending at this step
        """
        if profile:
            helperpk = profile.pk
        else:
            # Fake a profile that can be used in template
            profile = {"get_full_name": _("{Prénom} {Nom}")}
            helperpk = 0

        planning_link = self.request.build_absolute_uri(
            reverse(
                "season-availabilities-update",
                kwargs={"pk": self.season.pk, "helperpk": helperpk},
            )
        )

        return {
            "subject": settings.EMAIL_SUBJECT_PREFIX
            + gettext("Planning {season}").format(season=self.season.desc()),
            "body": {
                "pre": render_to_string(
                    "challenge/season_email_to_state_open_pre.txt",
                    {
                        "profile": profile,
                        "season": self.season,
                        "planning_link": planning_link,
                        "current_site": Site.objects.get_current(),
                    },
                ),
                "post": render_to_string(
                    "challenge/season_email_to_state_open_post.txt",
                    {
                        "profile": profile,
                        "season": self.season,
                        "planning_link": planning_link,
                        "current_site": Site.objects.get_current(),
                    },
                ),
            },
        }

    def dispatch(self, request, bypassperm=False, *args, **kwargs):
        """
        Push away if we're not allowed to go to that state now
        """
        if self.season and self.season.can_set_state_open:
            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied

    def get_email_recipients(self):
        """
        List of potential helpers/actors that we want to notify.
        This might include users that can't login.
        """
        return (
            self.potential_helpers_qs()
            .filter(profile__status__in=(USERSTATUS_ACTIVE, USERSTATUS_RESERVE))
            .exclude(Q(profile__formation="") & Q(profile__actor_for__isnull=True))
            .distinct()
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        recipients = list(
            self.get_email_recipients().order_by("first_name", "last_name", "id")
        )
        context["recipients"] = recipients
        context["some_recipients_cant_login"] = any(
            not user.profile.can_login for user in recipients
        )
        context["email"] = self.get_email()
        return context

    def form_valid(self, form):
        """
        Run our specific action here
        """
        # Save first
        form_result = super().form_valid(form)
        if form.cleaned_data["sendemail"]:
            # Then send emails
            for helper in self.get_email_recipients():
                email = self.get_email(helper)
                body = "\n".join(
                    [
                        email["body"]["pre"],
                        form.cleaned_data["customtext"],
                        email["body"]["post"],
                    ]
                )
                helper.profile.send_mail(email["subject"], body)
        return form_result


class SeasonAvailabilityReminderView(SeasonHelpersMixin, SeasonUpdateView):
    template_name = "challenge/season_availability_reminder.html"
    form_class = SeasonAvailabilityReminderForm

    def get_email(self, profile=None):
        if profile:
            helperpk = profile.pk
        else:
            profile = {"get_full_name": _("{Prénom} {Nom}")}
            helperpk = 0

        planning_link = self.request.build_absolute_uri(
            reverse(
                "season-availabilities-update",
                kwargs={"pk": self.season.pk, "helperpk": helperpk},
            )
        )

        pre = render_to_string(
            "challenge/season_email_availability_reminder.txt",
            {
                "profile": profile,
                "season": self.season,
                "planning_link": planning_link,
                "current_site": Site.objects.get_current(),
            },
        )

        return {
            "subject": settings.EMAIL_SUBJECT_PREFIX
            + gettext("Planning {season}").format(season=self.season.desc()),
            "body": {"pre": pre},
        }

    def get_email_recipients(self):
        base_qs = (
            self.potential_helpers_qs()
            .filter(profile__status__in=(USERSTATUS_ACTIVE, USERSTATUS_RESERVE))
            .exclude(Q(profile__formation="") & Q(profile__actor_for__isnull=True))
            .distinct()
        )
        sessions_qs = self.season.sessions_with_qualifs
        if not sessions_qs.exists():
            return base_qs.none()

        total_sessions = sessions_qs.count()

        # Helpers who have availabilities for all sessions (fully filled)
        fully_filled_helper_ids = (
            HelperSessionAvailability.objects.filter(session__in=sessions_qs)
            .values("helper_id")
            .annotate(n=Count("session", distinct=True))
            .filter(n=total_sessions)
            .values_list("helper_id", flat=True)
        )
        # Send to helpers not fully filled (partial or none)
        return base_qs.exclude(id__in=fully_filled_helper_ids)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        recipients = list(
            self.get_email_recipients().order_by("first_name", "last_name", "id")
        )
        context["recipients"] = recipients
        context["some_recipients_cant_login"] = any(
            not user.profile.can_login for user in recipients
        )
        context["email"] = self.get_email()
        return context

    def form_valid(self, form):
        form_result = super().form_valid(form)
        if not form.cleaned_data.get("sendemail"):
            return form_result
        now = timezone.now()
        for helper in self.get_email_recipients():
            email = self.get_email(helper)
            body = email["body"]["pre"]
            helper.profile.send_mail(email["subject"], body)

        self.object.availability_reminder_sent_at = now
        self.object.save()
        return form_result


class SeasonExportView(
    ExportMixin, SeasonAvailabilityMixin, HasPermissionsMixin, DetailView
):
    @property
    def export_filename(self):
        return _p("Singular month", "Mois") + "-" + "-".join(self.season.cantons)

    def undetected_translations(self):
        return [
            # Translators: Intervenant
            _("Int."),
            # Translators: Moniteur·trice + / Photographe
            _("M+"),
        ]

    def get_dataset(self):
        dataset = Dataset()
        # Prépare le fichier
        dataset.append_col(
            [
                gettext("Date"),
                gettext("Canton"),
                gettext("Établissement"),
                gettext("Emplacement"),
                gettext("Heures"),
                gettext("Nombre de qualifs"),
                # Logistique
                gettext("Moniteur·trice + / Photographe"),
                gettext("Mauvais temps"),
                gettext("Logistique vélos"),
                gettext("N° de contact vélos"),
                gettext("Pommes"),
                gettext("Total vélos"),
                gettext("Total casques"),
                gettext("Remarques"),
                # Qualif
                gettext("Classe"),
                gettext("Enseignant·e"),
                gettext("Moniteur·trice 2"),
                gettext("Moniteur·trice 1"),
                gettext("Moniteur·trice 1"),
                gettext("Nombre d’élèves"),
                gettext("Nombre de vélos"),
                gettext("Nombre de casques"),
                str(CATEGORY_CHOICE_A),
                str(CATEGORY_CHOICE_B),
                str(CATEGORY_CHOICE_C),
                gettext("Intervenant·e"),
                gettext("Remarques"),
            ]
        )
        for session in self.season.sessions:
            session_place = session.place
            if not session_place:
                session_place = (
                    session.address_city
                    if session.address_city
                    else session.orga.address_city
                )
            col = [
                date(session.day),
                session.orga.address_canton,
                session.orga.name,
                session_place,
                "%s - %s" % (time(session.begin), time(session.end)),
                session.n_qualifications,
                EXPORT_NAMETEL.format(
                    name=session.superleader.get_full_name(),
                    tel=session.superleader.profile.natel,
                )
                if session.superleader
                else "",
                str(session.fallback),
                session.bikes_concept,
                session.bikes_phone,
                session.apples,
                session.n_bikes,
                session.n_helmets,
                session.comments,
            ]
            if session.n_qualifications:
                for quali in session.qualifications.all():
                    if not len(col):
                        col = [""] * 14
                    col.append(quali.name)
                    col.append(
                        EXPORT_NAMETEL.format(
                            name=quali.class_teacher_fullname,
                            tel=quali.class_teacher_natel,
                        )
                        if quali.class_teacher_fullname
                        else ""
                    )
                    col.append(
                        EXPORT_NAMETEL.format(
                            name=quali.leader.get_full_name(),
                            tel=quali.leader.profile.natel,
                        )
                        if quali.leader
                        else ""
                    )
                    for i in range(2):
                        try:
                            helper = quali.helpers.all()[i]
                            col.append(
                                EXPORT_NAMETEL.format(
                                    name=helper.get_full_name(),
                                    tel=helper.profile.natel,
                                )
                                if helper
                                else ""
                            )
                        except IndexError:
                            col.append("")
                    col.append(quali.n_participants)
                    col.append(quali.n_bikes)
                    col.append(quali.n_helmets)
                    col.append(str(quali.activity_A) if quali.activity_A else "")
                    col.append(str(quali.activity_B) if quali.activity_B else "")
                    col.append(str(quali.activity_C) if quali.activity_C else "")
                    col.append(
                        EXPORT_NAMETEL.format(
                            name=quali.actor.get_full_name(),
                            tel=quali.actor.profile.natel,
                        )
                        if quali.actor
                        else ""
                    )
                    col.append(quali.comments)
                    dataset.append_col(col)
                    col = []
            else:
                col += [""] * 13
                dataset.append_col(col)
        return dataset


class GeneralPlanningSupportMixin(object):
    """
    Shared logic to support "general" planning (aggregated across months) for
    export views and feeds.
    """

    def _is_general_planning(self):
        return "year" in self.kwargs and "dv_season" in self.kwargs

    def _seasons_in_scope(self):
        try:
            year = int(self.kwargs.get("year"))
            dv_season = int(self.kwargs.get("dv_season"))
        except (TypeError, ValueError):
            return None, None, []
        qs = seasons_in_scope_for_user(self.request.user, year, dv_season)
        return year, dv_season, list(qs)

    def _helperpk_in_kwargs(self):
        helperpk_raw = self.kwargs.get("helperpk")
        try:
            helperpk = int(helperpk_raw) if helperpk_raw is not None else None
        except (TypeError, ValueError):
            helperpk = None
        return helperpk

    def _general_object(self):
        year, dv_season, seasons = self._seasons_in_scope()
        return GeneralSeason(
            year=year,
            dv_season=dv_season,
            seasons=seasons,
            helper_id=self._helperpk_in_kwargs(),
        )

    def _general_access_allowed(self, request):
        # Managers always
        permission_name = getattr(self, "required_permission", "challenge_season_crud")
        if has_permission(request.user, permission_name):
            return True
        # Helpers if any season is running
        _, _, seasons = self._seasons_in_scope()
        user_is_helper = request.user.profile.is_paid_staff and (
            request.user.profile.actor or request.user.profile.formation
        )
        any_running = any(s.staff_can_see_planning for s in seasons)
        if user_is_helper and any_running:
            return True
        return False

    def get_object(self, queryset=None):
        if self._is_general_planning():
            return self._general_object()
        return super().get_object(queryset)

    def dispatch(self, request, *args, **kwargs):
        if self._is_general_planning():
            if self._general_access_allowed(request):
                return super().dispatch(request, *args, **kwargs)
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)

    def custom_access_allowed(self, request):
        if self._is_general_planning():
            return self._general_access_allowed(request)
        return False


class SeasonPersonalPlanningExportView(
    GeneralPlanningSupportMixin, ExportMixin, SeasonAvailabilityMixin, DetailView
):
    allow_season_fetch = True
    raise_without_cantons = False
    view_is_planning = True

    @property
    def export_filename(self):
        # Support aggregated season (general planning)
        if hasattr(self.object, "dv_season"):
            label = _(slugify("Planning général"))
            return label + "-" + str(self.object.year)
        return _(slugify("Planning mois")) + "-" + "-".join(self.object.cantons)

    def get(self, request, *args, **kwargs):
        # Bypass SeasonAvailabilityMixin.get_context_data to avoid self.season usage in aggregated mode
        self.object = self.get_object()
        return self.render_to_response({})

    def get_dataset(self):
        dataset = Dataset()
        firstcol = [
            gettext("Date"),
            gettext("Canton"),
            gettext("Établissement"),
            gettext("Emplacement"),
            gettext("Heures"),
            gettext("Nombre de qualifs"),
        ]
        # Trouve toutes les personnes qui sont présentes dans ce mois / période
        qs = get_user_model().objects
        user_filters = [
            # Ceux qui ont répondu (quoi que ce soit)
            Q(availabilities__session__in=self.object.sessions_with_qualifs),
            # Moniteurs +
            Q(sess_monplus__in=self.object.sessions_with_qualifs),
            # Moniteurs 2
            Q(qualifs_mon2__session__in=self.object.sessions_with_qualifs),
            # Moniteurs 1
            Q(qualifs_mon1__session__in=self.object.sessions_with_qualifs),
            # Intervenants
            Q(qualifs_actor__session__in=self.object.sessions_with_qualifs),
        ]
        # Get users that match any filter in `user_filters`.
        # We used to do it in a single DB query with OR clauses,
        # but the production DB would execute this query in multiple minutes!
        user_ids = set()
        for user_filter in user_filters:
            user_ids.update(qs.filter(user_filter).values_list("id", flat=True))
        qs = (
            qs.filter(id__in=user_ids)
            .order_by("first_name", "last_name")
            .select_related("profile")
        )
        resolvermatch = self.request.resolver_match
        try:
            qs = qs.filter(pk=int(resolvermatch.kwargs["helperpk"]))
        except (KeyError, TypeError):
            pass

        firstcol += [user.get_full_name() for user in qs]
        dataset.append_col(firstcol)
        # Ajoute le canton d’affiliation comme deuxième colonne
        user_cantons_col = [""] * 6 + [user.profile.affiliation_canton for user in qs]
        dataset.append_col(user_cantons_col)
        # Limit sessions to those with a staff choice/assignment for the requested helper in aggregated mode
        sessions_qs = self.object.sessions_with_qualifs
        if hasattr(self.object, "dv_season"):
            resolvermatch = self.request.resolver_match
            try:
                helperpk = int(resolvermatch.kwargs["helperpk"])
            except (KeyError, TypeError):
                helperpk = None
            if helperpk:
                # Include sessions where the helper is assigned in any qualification
                chosen_ids = set(
                    sessions_qs.filter(
                        Q(qualifications__leader_id=helperpk)
                        | Q(qualifications__helpers__id=helperpk)
                        | Q(qualifications__actor_id=helperpk)
                    ).values_list("id", flat=True)
                )
                if chosen_ids:
                    sessions_qs = sessions_qs.filter(pk__in=chosen_ids)
                else:
                    sessions_qs = sessions_qs.none()

        for session in sessions_qs:
            session_place = session.place
            if not session_place:
                session_place = (
                    session.address_city
                    if session.address_city
                    else session.orga.address_city
                )
            col = [
                date(session.day),
                session.orga.address_canton,
                session.orga.name,
                session_place,
                "%s - %s" % (time(session.begin), time(session.end)),
                session.n_qualifications,
            ]
            users_roles = list(get_users_roles_for_session(qs, session).values())
            col += users_roles
            dataset.append_col(col)
        return dataset


class SeasonPlanningExportView(SeasonPersonalPlanningExportView, HasPermissionsMixin):
    view_is_planning = False


class SeasonAvailabilityView(SeasonAvailabilityMixin, DetailView):
    template_name = "challenge/season_availability.html"

    def get_context_data(self, **kwargs):
        context = super(SeasonAvailabilityView, self).get_context_data(**kwargs)
        # Add the form for picking a new helper
        context["form"] = SeasonNewHelperAvailabilityForm(cantons=self.season.cantons)
        hsas = self.current_availabilities()
        if hsas:
            # Fill in the helpers with the ones we currently have
            helpers_pks = self.current_availabilities().values_list(
                "helper_id", flat=True
            )
            potential_helpers = self.potential_helpers(
                qs=get_user_model().objects.filter(pk__in=helpers_pks)
            )

            context["potential_helpers"] = potential_helpers
            context["availabilities"] = self.get_initial(
                all_hsas=hsas, all_helpers=potential_helpers
            )
        return context

    def post(self, request, *args, **kwargs):
        seasonpk = kwargs.get("pk", None)
        form = SeasonNewHelperAvailabilityForm(request.POST)
        if form.is_valid():
            helper = form.cleaned_data["helper"]
            return HttpResponseRedirect(
                reverse_lazy(
                    "season-availabilities-update",
                    kwargs={"pk": seasonpk, "helperpk": helper.pk},
                )
            )
        return HttpResponseRedirect(
            reverse_lazy("season-availabilities", kwargs={"pk": seasonpk})
        )


class SeasonPlanningView(SeasonAvailabilityMixin, DetailView):
    template_name = "challenge/season_planning.html"
    allow_season_fetch = True
    raise_without_cantons = False
    view_is_planning = True

    def dispatch(self, request, *args, **kwargs):
        """
        Early request processing dispatch
        """
        try:
            """
            DEFIVELO-221 Circumvent a weird bug that made it so that emails were sent with helperpk == 0
            If helperpk == 0, redirect to self
            """
            if (
                self.request.user
                and int(self.request.resolver_match.kwargs["helperpk"]) == 0
            ):
                return HttpResponseRedirect(
                    reverse_lazy(
                        "season-planning",
                        kwargs={
                            "pk": self.season.pk,
                            "helperpk": self.request.user.pk,
                        },
                    )
                )

            """
            DEFIVELO-221 If we're in the wrong state, redirect to the other URL; move from planning to availability
            """
            if (
                self.request.user.profile.is_paid_staff
                and (
                    self.request.user.profile.actor
                    or self.request.user.profile.formation
                )
                and not self.season.staff_can_see_planning
                and self.season.staff_can_update_availability
            ):
                warning_message(
                    request,
                    _(
                        "Le planning n'est pas encore disponible; mets plutôt à jour tes disponibilités."
                    ),
                )
                return HttpResponseRedirect(
                    reverse_lazy(
                        "season-availabilities-update",
                        kwargs={
                            "pk": self.season.pk,
                            "helperpk": self.request.user.pk,
                        },
                    )
                )
        except KeyError:
            pass
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        potential_helpers = self.potential_helpers()
        context["submenu_category"] = "season-planning"
        context["potential_helpers"] = potential_helpers
        context["availabilities"] = (
            self.get_initial(all_helpers=potential_helpers) or {}
        )
        return context


class SeasonGeneralPlanningView(SeasonAvailabilityMixin, DetailView):
    template_name = "challenge/season_planning.html"
    allow_season_fetch = True
    raise_without_cantons = False
    view_is_planning = True

    def _seasons_in_scope(self):
        year = int(self.kwargs.get("year"))
        dv_season = int(self.kwargs.get("dv_season"))
        qs = seasons_in_scope_for_user(self.request.user, year, dv_season)
        return year, dv_season, list(qs)

    def get_object(self, queryset=None):
        year, dv_season, seasons = self._seasons_in_scope()
        helperpk_raw = self.kwargs.get("helperpk")
        try:
            helperpk = int(helperpk_raw) if helperpk_raw is not None else None
        except (TypeError, ValueError):
            helperpk = None
        return GeneralSeason(
            year=year, dv_season=dv_season, seasons=seasons, helper_id=helperpk
        )

    def dispatch(self, request, *args, **kwargs):
        helperpk_raw = self.kwargs.get("helperpk")
        if self.request.user and helperpk_raw is not None:
            try:
                if int(helperpk_raw) == 0:
                    year = int(self.kwargs.get("year"))
                    dv_season = int(self.kwargs.get("dv_season"))
                    return HttpResponseRedirect(
                        reverse_lazy(
                            "season-general-planning",
                            kwargs={
                                "year": year,
                                "dv_season": dv_season,
                                "helperpk": self.request.user.pk,
                            },
                        )
                    )
            except (TypeError, ValueError):
                pass

        # Authorization logic similar to SeasonPlanningView, adapted for aggregated seasons
        if has_permission(request.user, self.required_permission):
            return super().dispatch(request, *args, **kwargs)

        year, dv_season, seasons = self._seasons_in_scope()
        user_is_helper = request.user.profile.is_paid_staff and (
            request.user.profile.actor or request.user.profile.formation
        )
        any_running = any(s.staff_can_see_planning for s in seasons)
        any_open = any(s.staff_can_update_availability for s in seasons)
        if user_is_helper and any_running:
            return super().dispatch(request, *args, **kwargs)
        if user_is_helper and any_open:
            # Mirror the redirect used in SeasonPlanningView when planning isn't available
            target = next((s for s in seasons if s.staff_can_update_availability), None)
            if target:
                warning_message(
                    request,
                    _(
                        "Le planning n'est pas encore disponible. En attendant, vous pouvez mettre à jour vos disponibilités."
                    ),
                )
                return HttpResponseRedirect(
                    reverse_lazy(
                        "season-availabilities-update",
                        kwargs={"pk": target.pk, "helperpk": request.user.pk},
                    )
                )

        raise PermissionDenied

    def custom_access_allowed(self, request):
        # Allow aggregated general planning when appropriate; default rules handled in dispatch for redirects
        if has_permission(request.user, self.required_permission):
            return True
        year, dv_season, seasons = self._seasons_in_scope()
        user_is_helper = request.user.profile.is_paid_staff and (
            request.user.profile.actor or request.user.profile.formation
        )
        any_running = any(s.staff_can_see_planning for s in seasons)
        if user_is_helper and any_running:
            return True
        return False

    def get_context_data(self, **kwargs):
        context = DetailView.get_context_data(self, **kwargs)
        context["season"] = self.object

        helperpk_raw = self.kwargs.get("helperpk")
        try:
            helperpk = (
                int(helperpk_raw)
                if helperpk_raw is not None
                else int(self.request.user.pk)
            )
        except (TypeError, ValueError):
            helperpk = int(self.request.user.pk)

        potential_helpers = self.potential_helpers(
            qs=get_user_model().objects.filter(pk=helperpk)
        )
        context["submenu_category"] = "season-planning"
        context["helperpk"] = helperpk
        context["potential_helpers"] = potential_helpers
        context["availabilities"] = (
            self.get_initial(all_helpers=potential_helpers) or {}
        )
        if not context["availabilities"]:
            context["availabilities"] = OrderedDict()

        chosen_session_ids = set()
        choice_prefix = CHOICE_FIELDKEY.format(hpk=helperpk, spk="").replace("-s", "")
        for key, val in context["availabilities"].items():
            m = re.search(r"^" + re.escape(choice_prefix) + r"-s(\d+)$", key)
            if m and val:
                try:
                    chosen_session_ids.add(int(m.group(1)))
                except ValueError:
                    pass
        base_sessions = self.object.sessions_with_qualifs
        filtered_sessions = base_sessions.filter(pk__in=chosen_session_ids).annotate(
            n_qualifs=Count("qualifications", distinct=True),
            n_leaders=Count(
                Case(
                    When(
                        availability_statuses__chosen_as=CHOSEN_AS_LEADER,
                        then=F("availability_statuses__id"),
                    ),
                    output_field=IntegerField(),
                ),
                distinct=True,
            ),
            n_helpers=Count(
                Case(
                    When(
                        availability_statuses__chosen_as=CHOSEN_AS_HELPER,
                        then=F("availability_statuses__id"),
                    ),
                    output_field=IntegerField(),
                ),
                distinct=True,
            ),
            n_actors=Count(
                Case(
                    When(
                        availability_statuses__chosen_as=CHOSEN_AS_ACTOR,
                        then=F("availability_statuses__id"),
                    ),
                    output_field=IntegerField(),
                ),
                distinct=True,
            ),
        )
        context["sessions"] = filtered_sessions

        allowed_ids = set(filtered_sessions.values_list("pk", flat=True))
        pruned = OrderedDict()
        for key, value in context["availabilities"].items():
            m = re.search(r"-s(\d+)", key)
            if m:
                try:
                    spk = int(m.group(1))
                except ValueError:
                    spk = None
                if spk in allowed_ids:
                    # For availability entries, keep only if corresponding choice is True
                    if AVAILABILITY_FIELDKEY.format(hpk=helperpk, spk="") in key:
                        choice_key = CHOICE_FIELDKEY.format(hpk=helperpk, spk=spk)
                        if context["availabilities"].get(choice_key):
                            pruned[key] = value
                    else:
                        pruned[key] = value
            else:
                pruned[key] = value
        context["availabilities"] = pruned

        _, _, seasons = self._seasons_in_scope()
        context["user_can_see_season"] = any(
            s.unprivileged_user_can_see(self.request.user) for s in seasons
        )
        return context


class SeasonAvailabilityUpdateView(SeasonAvailabilityMixin, SeasonUpdateView):
    template_name = "challenge/season_availability_update.html"
    success_message = _("Disponibilités mises à jour")
    form_class = SeasonAvailabilityForm
    allow_season_fetch = True
    raise_without_cantons = False
    view_is_update = True

    def dispatch(self, request, *args, **kwargs):
        """
        Early request processing dispatch
        """
        try:
            """
            DEFIVELO-221 Circumvent a weird bug that made it so that emails were sent with helperpk == 0
            If helperpk == 0, redirect to self
            """
            if (
                self.request.user
                and int(self.request.resolver_match.kwargs["helperpk"]) == 0
            ):
                return HttpResponseRedirect(
                    reverse_lazy(
                        "season-availabilities-update",
                        kwargs={
                            "pk": self.season.pk,
                            "helperpk": self.request.user.pk,
                        },
                    )
                )
            """
            DEFIVELO-221 If we're in the wrong state, redirect to the other URL; move from planning to availability
            """
            if (
                # Poweruser & Related state manager, can still access
                not self.request.user.managedstates.filter(
                    canton__in=self.season.cantons
                ).exists()
                and not has_role(self.request.user, "power_user")
                and (
                    self.request.user.profile.actor
                    or self.request.user.profile.formation
                )
                and not self.season.staff_can_update_availability
                and self.season.staff_can_see_planning
            ):
                warning_message(
                    request,
                    _(
                        "Il n'est pas possible de mettre à jour tes disponibilités; voici le planning."
                    ),
                )
                return HttpResponseRedirect(
                    reverse_lazy(
                        "season-planning",
                        kwargs={
                            "pk": self.season.pk,
                            "helperpk": self.request.user.pk,
                        },
                    )
                )
        except KeyError:
            pass
        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        form_kwargs = super(SeasonAvailabilityUpdateView, self).get_form_kwargs()
        form_kwargs["potential_helpers"] = self.potential_helpers()
        return form_kwargs

    def get_context_data(self, **kwargs):
        context = super(SeasonAvailabilityUpdateView, self).get_context_data(**kwargs)
        context["potential_helpers"] = self.potential_helpers()
        return context

    def get_success_url(self):
        try:
            usercantons = user_cantons(self.request.user)
        except LookupError:
            usercantons = None
        if has_permission(self.request.user, "challenge_season_crud") and list(
            set(usercantons).intersection(set(self.season.cantons))
        ):
            return reverse_lazy("season-availabilities", kwargs={"pk": self.season.pk})
        else:
            return reverse_lazy(
                "season-availabilities-update",
                kwargs={
                    "pk": self.season.pk,
                    "helperpk": self.request.user.pk,
                },
            )

    def form_valid(self, form):
        # Create or update the Availability objects
        for session in self.object.sessions_with_qualifs:
            for helper_category, helpers in self.potential_helpers():
                for helper in helpers:
                    # Fill in the wishes
                    workwishkey = SEASON_WORKWISH_FIELDKEY.format(hpk=helper.pk)
                    if workwishkey in form.cleaned_data:
                        amount = form.cleaned_data[workwishkey]
                        try:
                            (hww, created) = HelperSeasonWorkWish.objects.get_or_create(
                                season=self.season,
                                helper=helper,
                                defaults={"amount": amount},
                            )
                        except HelperSeasonWorkWish.MultipleObjectsReturned:
                            # Too many of these, for some reason. Take the latest.
                            hww = HelperSeasonWorkWish.objects.filter(
                                season=self.season, helper=helper
                            ).order_by("-id")[0]
                            created = False
                        if not created:
                            hww.amount = amount
                            hww.save()

                    # Fill in the availabilities
                    fieldkey = AVAILABILITY_FIELDKEY.format(
                        hpk=helper.pk, spk=session.pk
                    )
                    if fieldkey in form.cleaned_data:
                        availability = form.cleaned_data[fieldkey]
                        if availability:
                            (
                                hsa,
                                created,
                            ) = HelperSessionAvailability.objects.get_or_create(
                                session=session,
                                helper=helper,
                                defaults={"availability": availability},
                            )
                            if not created:
                                hsa.availability = availability
                                hsa.save()
        return super(SeasonAvailabilityUpdateView, self).form_valid(form)


class SeasonStaffChoiceUpdateView(
    SeasonAvailabilityMixin, SeasonUpdateView, HasPermissionsMixin
):
    template_name = "challenge/season_staff_update.html"
    success_message = _("Choix du personnel mises à jour")
    form_class = SeasonStaffChoiceForm
    form_filter_class = SeasonStaffFilterForm
    view_is_update = True

    def get_initial(self):
        # Shortcut through giving only the available_helpers
        return super(SeasonStaffChoiceUpdateView, self).get_initial(
            all_hsas=None, all_helpers=self.available_helpers
        )

    def get_form_kwargs(self):
        form_kwargs = super(SeasonStaffChoiceUpdateView, self).get_form_kwargs()
        form_kwargs["available_helpers"] = self.available_helpers
        return form_kwargs

    def get_context_data(self, **kwargs):
        context = super(SeasonStaffChoiceUpdateView, self).get_context_data(**kwargs)
        context["available_helpers"] = self.available_helpers
        context["season_staff_filter_form"] = SeasonStaffFilterForm(
            organisations=self._extract_orgas(context["sessions"])
        )
        context["availabilities"] = self.get_initial()
        return context

    def form_valid(self, form):
        # Update all staff choices
        for session in self.object.sessions_with_qualifs:
            session_helpers = {}  # Those picked for the season
            session_non_helpers = {}  # Those not picked for the season
            for helper_category, helpers in self.available_helpers:
                for helper in helpers:
                    fieldkey = STAFF_FIELDKEY.format(hpk=helper.pk, spk=session.pk)

                    try:
                        chosen_as = int(form.cleaned_data[fieldkey])

                        HelperSessionAvailability.objects.filter(
                            session=session, helper=helper
                        ).update(chosen_as=chosen_as)
                    except ValueError:
                        chosen_as = None

                    if chosen_as != CHOSEN_AS_NOT:
                        session_helpers[helper.pk] = helper
                    else:
                        session_non_helpers[helper.pk] = helper

            # Do a session-wide check across all helpers picked for that
            # session
            for quali in session.qualifications.all():
                for non_helper in session_non_helpers.values():
                    # Drop those not in the session anymore
                    if non_helper == quali.leader:
                        quali.leader = None
                    if non_helper in quali.helpers.all():
                        quali.helpers.remove(helper)
                    if non_helper == quali.actor:
                        quali.actor = None
                quali.save()
        return HttpResponseRedirect(
            reverse_lazy("season-availabilities", kwargs={"pk": self.object.pk})
        )

    def _extract_orgas(self, sessions):
        seen_ids = set()
        orgas = []
        for session in sessions:
            if session.orga and session.orga_id not in seen_ids:
                seen_ids.add(session.orga_id)
                orgas.append(session.orga)
        return orgas


class SeasonCreateView(
    SeasonMixin, HasPermissionsMixin, SuccessMessageMixin, CreateView
):
    success_message = _("Mois créé")


class SeasonHelperListView(HelpersList, HasPermissionsMixin, SeasonMixin):
    model = get_user_model()
    page_title = _("Moniteur·trice·s du mois")

    def get_context_data(self, **kwargs):
        context = super(SeasonHelperListView, self).get_context_data(**kwargs)
        # Add our submenu_category context
        context["submenu_category"] = "season-helperlist"
        return context

    def get_queryset(self):
        return (
            super(SeasonHelperListView, self)
            .get_queryset()
            .filter(
                Q(qualifs_mon2__session__in=self.season.sessions_with_qualifs)
                | Q(qualifs_mon1__session__in=self.season.sessions_with_qualifs),
            )
            .distinct()
        )


class SeasonActorListView(ActorsList, HasPermissionsMixin, SeasonMixin):
    model = get_user_model()
    page_title = _("Intervenant·e·s du mois")

    def get_context_data(self, **kwargs):
        context = super(SeasonActorListView, self).get_context_data(**kwargs)
        # Add our submenu_category context
        context["submenu_category"] = "season-actorslist"
        return context

    def get_queryset(self):
        return (
            super(SeasonActorListView, self)
            .get_queryset()
            .filter(qualifs_actor__session__in=self.season.sessions_with_qualifs)
            .distinct()
        )


class SeasonErrorsListView(HasPermissionsMixin, SeasonMixin, ListView):
    required_permission = "challenge_season_crud"
    model = Qualification
    page_title = _("Erreurs dans les qualifs du mois")
    template_name = "challenge/season_errors.html"
    context_object_name = "qualifs"

    def get_context_data(self, **kwargs):
        context = super(SeasonErrorsListView, self).get_context_data(**kwargs)
        # Add our submenu_category context
        context["submenu_category"] = "season-errorslist"
        return context

    def get_queryset(self):
        wrong_qualifs = []
        for session in self.season.sessions_with_qualifs.all():
            for quali in session.qualifications.all():
                if quali.has_availability_incoherences:
                    wrong_qualifs.append(quali.pk)
        return (
            super(SeasonErrorsListView, self)
            .get_queryset()
            .filter(pk__in=wrong_qualifs)
            .distinct()
        )


class SeasonPersonalPlanningExportFeed(GeneralPlanningSupportMixin, ICalFeed):
    timezone = datetime.timezone.utc
    file_name = "sessions.ics"

    def __call__(self, request, *args, **kwargs):
        self.request = request
        self.kwargs = kwargs
        resolvermatch = self.request.resolver_match
        self.user = get_user_model().objects.get(
            pk=int(resolvermatch.kwargs["helperpk"])
        )

        # Support "general" planning scope via the shared mixin
        if self._is_general_planning():
            if not self._general_access_allowed(request):
                raise PermissionDenied
            self.object = self.season = self._general_object()
        else:
            self.object = self.season = Season.objects.get(pk=kwargs.get("pk"))
        return super().__call__(request)

    def items(self):
        return self.object.sessions_with_qualifs.filter(
            Q(superleader=self.user)
            | Q(qualifications__actor=self.user)
            | Q(qualifications__helpers=self.user)
            | Q(qualifications__leader=self.user)
        )

    def item_guid(self, session):
        return "{}-{}".format(session.id, "session")

    def item_title(self, session):
        roles_by_user = get_users_roles_for_session([self.user], session)
        role_label = roles_by_user[self.user]
        return " ".join([role_label, session.orga.address_canton, session.orga.name])

    def item_description(self, session):
        session_place = session.place
        if not session_place:
            session_place = (
                session.address_city
                if session.address_city
                else session.orga.address_city
            )
        return " ".join(
            [
                session_place,
                "%s - %s" % (time(session.begin), time(session.end)),
            ]
        )

    def item_start_datetime(self, session):
        return session.start_datetime.astimezone(datetime.timezone.utc)

    def item_end_datetime(self, session):
        return session.end_datetime.astimezone(datetime.timezone.utc)

    def item_link(self, session):
        # Link back to the appropriate planning page depending on mode
        if hasattr(self.object, "dv_season"):
            return reverse(
                "season-general-planning",
                kwargs={
                    "year": self.object.year,
                    "dv_season": self.object.dv_season,
                    "helperpk": self.user.id,
                },
            )
        return reverse(
            "season-planning", kwargs={"pk": self.season.pk, "helperpk": self.user.id}
        )



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0009_userprofile_address_3"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="activity_cantons",
            field=models.CharField(
                max_length=77,
                blank=True,
                verbose_name="Cantons d'affiliation",
                choices=[
                    ("AG", "Aargau"),
                    ("AI", "Appenzell Innerrhoden"),
                    ("AR", "Appenzell Ausserrhoden"),
                    ("BS", "Basel-Stadt"),
                    ("BL", "Basel-Land"),
                    ("BE", "Berne"),
                    ("FR", "Fribourg"),
                    ("GE", "Geneva"),
                    ("GL", "Glarus"),
                    ("GR", "Graubuenden"),
                    ("JU", "Jura"),
                    ("LU", "Lucerne"),
                    ("NE", "Neuchatel"),
                    ("NW", "Nidwalden"),
                    ("OW", "Obwalden"),
                    ("SH", "Schaffhausen"),
                    ("SZ", "Schwyz"),
                    ("SO", "Solothurn"),
                    ("SG", "St. Gallen"),
                    ("TG", "Thurgau"),
                    ("TI", "Ticino"),
                    ("UR", "Uri"),
                    ("VS", "Valais"),
                    ("VD", "Vaud"),
                    ("ZG", "Zug"),
                    ("ZH", "Zurich"),
                ],
            ),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="bagstatus",
            field=models.PositiveSmallIntegerField(
                default=0,
                verbose_name="Statut",
                choices=[
                    (0, "---------"),
                    (10, "Actif"),
                    (20, "Réserve"),
                    (30, "Inactif"),
                    (30, "Archive"),
                ],
            ),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="bagstatus_updatetime",
            field=models.DateTimeField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="comments",
            field=models.TextField(blank=True, verbose_name="Remarques"),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="firstmed_course",
            field=models.BooleanField(default=False, verbose_name="Cours samaritains"),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="firstmed_course_comm",
            field=models.CharField(
                max_length=255, blank=True, verbose_name="Cours samaritains (spécifier)"
            ),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="pedagogical_experience",
            field=models.TextField(blank=True, verbose_name="Expérience pédagogique"),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="status",
            field=models.PositiveSmallIntegerField(
                default=0,
                verbose_name="Statut",
                choices=[
                    (0, "---------"),
                    (10, "Actif"),
                    (20, "Réserve"),
                    (30, "Inactif"),
                    (30, "Archive"),
                ],
            ),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="status_updatetime",
            field=models.DateTimeField(null=True, blank=True),
        ),
    ]



import datetime

from django.core import mail
from django.urls import reverse

from apps.challenge.tests.factories import QualificationFactory, SessionFactory
from apps.common import DV_STATES
from apps.orga.tests.factories import OrganizationFactory
from apps.user.tests.factories import UserFactory
from defivelo.roles import user_cantons
from defivelo.tests.utils import AuthClient, StateManagerAuthClient

from .factories import TimesheetFactory, ValidatedTimesheetFactory


def test_reminder_preview_loads(db):
    client = StateManagerAuthClient()
    managed_cantons = user_cantons(client.user)

    QualificationFactory(
        actor=UserFactory(
            first_name="Maurice",
            last_name="Moss",
            profile__affiliation_canton=managed_cantons[0],
        ),
        session=SessionFactory(
            day=datetime.date(2019, 4, 11),
            orga=OrganizationFactory(address_canton=managed_cantons[0]),
        ),
    )

    response = client.get(
        reverse("salary:send-timesheets-reminder", kwargs={"year": 2019, "month": 4})
    )
    assert response.status_code == 200


def test_reminder_is_sent_to_helpers_with_missing_timesheets(db):
    client = StateManagerAuthClient()
    managed_cantons = user_cantons(client.user)

    # Helpers with all timesheets
    helper1 = UserFactory(profile__affiliation_canton=managed_cantons[0])
    helper2 = UserFactory(profile__affiliation_canton=managed_cantons[0])
    # Helper with missing timesheet
    helper3 = UserFactory(profile__affiliation_canton=managed_cantons[0])

    date1 = datetime.date(2019, 2, 11)
    date2 = datetime.date(2019, 2, 12)
    orga = OrganizationFactory(address_canton=managed_cantons[0])
    session1 = SessionFactory(day=date1, orga=orga)
    session2 = SessionFactory(day=date2, orga=orga)

    QualificationFactory(actor=helper1, session=session1)
    QualificationFactory(actor=helper1, session=session2)
    TimesheetFactory(user=helper1, date=date1)
    ValidatedTimesheetFactory(user=helper1, date=date2)

    QualificationFactory(actor=helper2, session=session1)
    TimesheetFactory(user=helper2, date=date1)

    QualificationFactory(actor=helper3, session=session1)
    QualificationFactory(actor=helper3, session=session2)
    TimesheetFactory(user=helper3, date=date1)

    mail.outbox = []
    response = client.post(
        reverse("salary:send-timesheets-reminder", kwargs={"year": 2019, "month": 2})
    )
    assert response.status_code == 302
    assert len(mail.outbox) == 1
    assert mail.outbox[0].recipients() == [helper3.email]
    assert "{Prénom} {Nom}" not in mail.outbox[0].body


def test_reminder_is_not_sent_to_helper_of_other_canton(db):
    client = StateManagerAuthClient()

    assert DV_STATES[1] not in client.user.profile.managed_cantons  # invariant
    QualificationFactory(
        actor=UserFactory(profile__affiliation_canton=DV_STATES[1]),
        session=SessionFactory(
            day=datetime.date(2019, 2, 11),
            orga=OrganizationFactory(address_canton=DV_STATES[1]),
        ),
    )

    mail.outbox = []
    response = client.post(
        reverse("salary:send-timesheets-reminder", kwargs={"year": 2019, "month": 2})
    )
    assert response.status_code == 302
    assert len(mail.outbox) == 0


def test_reminder_is_not_sent_to_helper_of_other_month(db):
    client = StateManagerAuthClient()
    managed_cantons = user_cantons(client.user)

    QualificationFactory(
        actor=UserFactory(profile__affiliation_canton=managed_cantons[0]),
        session=SessionFactory(day=datetime.date(2019, 1, 11)),
    )

    mail.outbox = []
    response = client.post(
        reverse("salary:send-timesheets-reminder", kwargs={"year": 2019, "month": 2})
    )
    assert response.status_code == 302
    assert len(mail.outbox) == 0


def test_helper_cant_send_reminder(db):
    client = AuthClient()
    response = client.post(
        reverse("salary:send-timesheets-reminder", kwargs={"year": 2019, "month": 2})
    )
    assert response.status_code == 403



from django.apps import AppConfig


class UserConfig(AppConfig):
    name = "apps.user"
    verbose_name = "Users"

    def ready(self) -> None:  # type: ignore[override]
        # Import signal handlers to ensure they are registered at startup
        from . import signals  # noqa: F401

        return super().ready()



from datetime import timedelta


def _get_qualifs_calendar_structured_data(our_sessions):
    """
    Alternative to mixins and parent class mess
    """
    date_sessions = []

    # Get to its monday (see https://stackoverflow.com/questions/1622038/find-mondays-date-with-python)
    first_session_day = our_sessions[0].day
    first_monday = first_session_day + timedelta(days=-first_session_day.weekday())
    thisday = first_monday
    offset = 0
    while len(our_sessions) > 0:
        thisday = first_monday + timedelta(days=offset)
        struct = {"day": thisday, "sessions": []}
        while our_sessions and our_sessions[0].day == thisday:
            struct["sessions"].append(our_sessions.pop(0))
        date_sessions.append(struct)
        # Go to next day
        offset = offset + 1
    # Fill in the missing days
    while thisday.weekday() != 6:
        thisday = first_monday + timedelta(days=offset)
        date_sessions.append({"day": thisday, "sessions": []})
        offset = offset + 1
    return date_sessions



# Generated by Django 2.2.24 on 2022-01-20 09:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0069_add_zoug"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2016 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.http import HttpResponse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from import_export.formats import base_formats


class ExportMixin(object):
    export_class = None
    export_filename = None
    export_kwargs = {}

    def render_to_response(self, context, **response_kwargs):
        resolvermatch = self.request.resolver_match
        formattxt = resolvermatch.kwargs.get("format", "csv")
        # Instantiate the format object from base_formats in import_export
        try:
            format = getattr(base_formats, formattxt.upper())()
        except AttributeError:
            format = base_formats.CSV()

        try:
            dataset = (self.get_export_class(self.request)).export(self.object_list)
        except AttributeError:
            dataset = self.get_dataset()

        filename = _("DV-{exportfilename}_{YMD_date}.{extension}").format(
            exportfilename=self.export_filename,
            YMD_date=timezone.now().strftime("%Y%m%d"),
            extension=format.get_extension(),
        )

        response = HttpResponse(
            dataset.export(formattxt, **self.export_kwargs),
            format.get_content_type() + ";charset=utf-8",
        )
        response["Content-Disposition"] = 'attachment; filename="{f}"'.format(
            f=filename
        )
        return response

    def get_export_class(self, request):
        """
        Allow a dynamic export_class setting
        """
        return self.export_class


class PaginatorMixin(object):
    paginate_by = 10
    paginate_orphans = 3

    def get_context_data(self, *args, **kwargs):
        context = super(PaginatorMixin, self).get_context_data(*args, **kwargs)
        # Re-create the filtered querystring from GET, drop page off it
        querydict = self.request.GET.copy()
        try:
            del querydict["page"]
        except Exception:
            pass
        context["filter_querystring"] = querydict.urlencode()
        return context



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from django import forms
from django.contrib import admin

from allauth.account.admin import EmailAddressAdmin as BaseEmailAddressAdmin
from allauth.account.models import EmailAddress

from .models import UserManagedState, UserProfile


class AdminUserProfile(admin.ModelAdmin):
    list_filter = ["updated_at"]


admin.site.register(UserProfile, AdminUserProfile)
admin.site.register(UserManagedState)


class EmailAddressAdminForm(forms.ModelForm):
    class Meta:
        model = EmailAddress
        fields = "__all__"

    def save(self, commit=True):
        if self.cleaned_data["primary"] is True:
            EmailAddress.objects.filter(
                user=self.instance.user,
                primary=True,
            ).exclude(id=self.instance.id).update(primary=False)

        return super().save(commit)


class EmailAddressAdmin(BaseEmailAddressAdmin):
    list_filter = ("primary",)
    form = EmailAddressAdminForm


admin.site.unregister(EmailAddress)
admin.site.register(EmailAddress, EmailAddressAdmin)



# Generated by Django 2.2.16 on 2020-12-14 16:04

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0076_registration"),
    ]

    operations = [
        migrations.AddField(
            model_name="registration",
            name="is_archived",
            field=models.BooleanField(default=False, verbose_name="Archivée"),
        ),
        migrations.AlterField(
            model_name="registration",
            name="day_time",
            field=models.CharField(
                choices=[("08:30:00", "Matin"), ("13:30:00", "Après-midi")],
                default="08:30:00",
                max_length=8,
                verbose_name="Moment de la journée",
            ),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0032_session_superleader"),
    ]

    operations = [
        migrations.AlterField(
            model_name="season",
            name="cantons",
            field=models.CharField(
                max_length=38,
                verbose_name="Cantons",
                choices=[
                    ("BS", "Basel-Stadt"),
                    ("BE", "Berne"),
                    ("FR", "Fribourg"),
                    ("GE", "Geneva"),
                    ("JU", "Jura"),
                    ("LU", "Lucerne"),
                    ("NE", "Neuchatel"),
                    ("SG", "St. Gallen"),
                    ("VS", "Valais"),
                    ("VD", "Vaud"),
                    ("ZH", "Zurich"),
                    ("VS-OW", "Haut-Valais"),
                ],
            ),
        ),
    ]



# Generated by Django 2.2.9 on 2020-03-06 09:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("salary", "0002_timesheet_validated_by"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="timesheet",
            name="time_monitor",
        ),
        migrations.AddField(
            model_name="timesheet",
            name="time_helper",
            field=models.FloatField(default=0, verbose_name="Heures moni·teur·trice"),
        ),
        migrations.AlterField(
            model_name="timesheet",
            name="time_actor",
            field=models.FloatField(default=0, verbose_name="Heures intervenant·e"),
        ),
    ]



# Generated by Django 3.2.13 on 2022-04-19 07:57
from django.db import migrations, models
from django.db.models import F, Value
from django.db.models.functions import Concat

import apps.common.fields


def forward(apps, schema_editor):
    """
    Reformat data for the new FieldType
    """
    UserProfile = apps.get_model("user", "UserProfile")
    UserProfile.objects.update(
        languages_challenges=Concat(Value("{"), F("languages_challenges"), Value("}"))
    )


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0071_userprofile_activity_cantons"),
    ]

    operations = [
        migrations.RunPython(forward),
        migrations.AlterField(
            model_name="userprofile",
            name="languages_challenges",
            field=apps.common.fields.ChoiceArrayField(
                base_field=models.CharField(
                    choices=[
                        ("fr", "French"),
                        ("de", "German"),
                        ("it", "Italian"),
                        ("en", "English"),
                    ],
                    max_length=2,
                ),
                blank=True,
                default=list,
                size=None,
                verbose_name="Prêt à animer en",
            ),
        ),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0030_auto_20161010_1519"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="qualification",
            options={
                "verbose_name": "Qualif'",
                "verbose_name_plural": "Qualifs",
                "ordering": ["session", "created_on", "name"],
            },
        ),
        migrations.AlterField(
            model_name="qualification",
            name="activity_A",
            field=models.ForeignKey(
                verbose_name="Vélo urbain",
                to="challenge.QualificationActivity",
                blank=True,
                null=True,
                related_name="qualifs_A",
                on_delete=models.CASCADE,
            ),
        ),
        migrations.AlterField(
            model_name="qualificationactivity",
            name="category",
            field=models.CharField(
                max_length=1,
                verbose_name="Catégorie",
                choices=[("A", "Vélo urbain"), ("B", "Mécanique"), ("C", "Rencontre")],
                blank=True,
            ),
        ),
    ]



import datetime
from datetime import time
from typing import Mapping

from django.contrib.auth import get_user_model
from django.db.models import Count, Q, Sum
from django.utils.safestring import mark_safe
from django.utils.translation import gettext
from django.utils.translation import gettext as _

from apps.common import DV_SEASON_AUTUMN, DV_SEASON_LAST_SPRING_MONTH, DV_SEASON_SPRING
from apps.user import FORMATION_M1, FORMATION_M2, formation_short

from . import (
    CHOSEN_AS_ACTOR,
    CHOSEN_AS_HELPER,
    CHOSEN_AS_LEADER,
    CHOSEN_AS_NOT,
    CHOSEN_AS_REPLACEMENT,
)

User = get_user_model()


def get_users_roles_for_session(users, session) -> Mapping[User, str]:
    """
    Get the role (in short format) of each user of `users` in `session`.
    """
    user_session_chosen_as = dict(
        session.availability_statuses.exclude(chosen_as=CHOSEN_AS_NOT).values_list(
            "helper_id", "chosen_as"
        )
    )
    qualifications = session.qualifications.all()

    roles = {}
    for user in users:
        label = ""
        for quali in qualifications:
            if user == quali.leader:
                label = formation_short(FORMATION_M2, True)
                break
            elif user in quali.helpers.all():
                label = formation_short(FORMATION_M1, True)
                break
            elif user == quali.actor:
                # Translators: Nom court pour 'Intervenant'
                label = gettext("Int.")
                break
        # Vérifie tout de même si l’utilisateur est déjà sélectionné
        if not label and user.id in user_session_chosen_as:
            if user_session_chosen_as[user.id] == CHOSEN_AS_LEADER:
                label = formation_short(FORMATION_M2, True)
            elif user_session_chosen_as[user.id] == CHOSEN_AS_HELPER:
                label = formation_short(FORMATION_M1, True)
            elif user_session_chosen_as[user.id] == CHOSEN_AS_ACTOR:
                # Translators: Nom court pour 'Intervenant'
                label = gettext("Int.")
            elif user_session_chosen_as[user.id] == CHOSEN_AS_REPLACEMENT:
                # Translators: Nom court pour 'Moniteur·trice de secours'
                label = gettext("S")
            else:
                label = gettext("×")

        if user == session.superleader:
            # Translators: Nom court pour 'Moniteur +'
            label = "{} / {}".format(label, gettext("M+")) if label else gettext("M+")

        roles[user] = label
    return roles


def is_morning(begin):
    return begin <= time(12, 00)


class GeneralSeason(object):
    """
    Aggregated planning wrapper over multiple `Season` instances for a year and DV season.
    Provides computed date range, cantons and session helpers; used by general planning views/exports.
    """

    def __init__(self, *, year, dv_season, seasons, helper_id=None):
        from calendar import monthrange

        self.year = year
        self.dv_season = dv_season
        self._seasons = seasons
        self.helper_id = helper_id
        self.pk = seasons[0].pk if seasons else None

        if dv_season == DV_SEASON_SPRING:
            self.begin = datetime.date(year, 1, 1)
            __, last_day = monthrange(year, DV_SEASON_LAST_SPRING_MONTH)
            self.end = datetime.date(year, DV_SEASON_LAST_SPRING_MONTH, last_day)
            self.season_full = _("Printemps {year}").format(year=year)
        elif DV_SEASON_AUTUMN:
            self.begin = datetime.date(year, DV_SEASON_LAST_SPRING_MONTH + 1, 1)
            self.end = datetime.date(year, 12, 31)
            self.season_full = _("Automne {year}").format(year=year)
        else:
            raise Exception(_("Pas de saison valide pour cette période"))

        cantons = get_cantons_for_seasons(seasons)

        self._base_cantons = cantons
        self.cantons = list(cantons)

        self.state_icon = ""
        self.state_full = _("Planning général")

        self._sessions_with_q = None

    @property
    def sessions_with_qualifs(self):
        from .models import Session

        if self._sessions_with_q is None:
            qs = (
                Session.objects.filter(
                    orga__address_canton__in=self._base_cantons,
                    day__gte=self.begin,
                    day__lte=self.end,
                )
                .prefetch_related(
                    "availability_statuses",
                    "qualifications",
                    "qualifications__leader",
                    "qualifications__helpers",
                    "qualifications__actor",
                    "orga",
                )
                .annotate(Count("qualifications"))
                .filter(qualifications__count__gt=0)
                .order_by("day", "begin", "orga__name")
            )
            # If a helper is specified, restrict to sessions relevant to that helper
            if self.helper_id:
                qs = qs.filter(
                    Q(superleader_id=self.helper_id)
                    | (
                        Q(availability_statuses__helper_id=self.helper_id)
                        & ~Q(availability_statuses__chosen_as=CHOSEN_AS_NOT)
                    )
                    | Q(qualifications__leader_id=self.helper_id)
                    | Q(qualifications__helpers__id=self.helper_id)
                    | Q(qualifications__actor_id=self.helper_id)
                ).distinct()
                # Update cantons to only those with confirmed availability/assignment
                self.cantons = list(
                    qs.values_list("orga__address_canton", flat=True).distinct()
                )
            self._sessions_with_q = qs
        return self._sessions_with_q

    @property
    def work_wishes(self):
        from .models.availability import HelperSeasonWorkWish

        # Aggregate work wishes across all seasons by helper
        return (
            HelperSeasonWorkWish.objects.filter(season__in=self._seasons)
            .values("helper_id")
            .annotate(amount=Sum("amount"))
        )

    def desc(self, abbr=False):
        from defivelo.templatetags.dv_filters import cantons_abbr

        cantons_str = ", ".join(cantons_abbr(self.cantons, abbr))
        return mark_safe(
            _("{cantons} - {season_full}").format(
                cantons=cantons_str, season_full=self.season_full
            )
        )

    @property
    def desc_abbr(self):
        return self.desc(True)


def seasons_in_scope_for_user(user, year, dv_season):
    """
    Return queryset of seasons for a given user/year/dv_season using the same
    canton scoping as SeasonMixin/UserProfile.get_seasons.
    """
    qs = user.profile.get_seasons(False).filter(year=year)
    if dv_season == DV_SEASON_SPRING:
        qs = qs.filter(month_start__lte=DV_SEASON_LAST_SPRING_MONTH)
    elif dv_season == DV_SEASON_AUTUMN:
        qs = qs.filter(month_start__gt=DV_SEASON_LAST_SPRING_MONTH)
    else:
        qs = qs.none()
    return qs


def get_cantons_for_seasons(seasons):
    """Return unique list of cantons across the given seasons."""
    cantons = []
    for s in seasons:
        for c in s.cantons:
            if c not in cantons:
                cantons.append(c)
    return cantons



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015-2020 Didier Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.template.defaultfilters import date
from django.utils.translation import gettext_lazy as _

from localflavor.ch.forms import CHStateSelect

from apps.common.forms import (
    CHPhoneNumberField,
    SwissDateField,
    SwissTimeField,
    UserAutoComplete,
)
from apps.user import STATE_CHOICES_WITH_DEFAULT
from apps.user.models import USERSTATUS_DELETED

from ..models import Session

nodate_change_warning = _(
    "La date ne peut plus être modifiée, car des heures ont déjà été "
    "saisies dans au moins une des Qualifs de cette session."
)


class SessionForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        kwargs.pop("cantons", None)
        self.season = kwargs.pop("season", None)
        is_for_coordinator = kwargs.pop("is_for_coordinator", False)

        super().__init__(**kwargs)
        if self.season.cantons:
            # Only permit orgas within the allowed cantons
            qs = self.fields["orga"].queryset.filter(
                address_canton__in=self.season.cantons
            )
            self.fields["orga"].queryset = qs

        # Disable minDate and maxDate - DEFIVELO-98
        # try:
        #     self.fields["day"].widget.options["minDate"] = self.season.begin.strftime(
        #         "%Y-%m-%d"
        #     )
        #     self.fields["day"].widget.options["maxDate"] = self.season.end.strftime(
        #         "%Y-%m-%d"
        #     )
        # except Exception:
        #     pass

        if is_for_coordinator:
            # For non-stateManagers (coordinator), set some fields disabled
            for f in ["orga", "day"]:
                self.fields[f].widget.attrs["readonly"] = True
                self.fields[f].disabled = True
            # And lots as hidden
            for f in list(self.fields.keys()):
                if f not in [
                    "orga",
                    "day",
                    "begin",
                    "fallback_plan",
                    "place",
                    "address_street",
                    "address_no",
                    "address_zip",
                    "address_city",
                    "address_canton",
                ]:
                    del self.fields[f]

        if self.instance.pk and self.instance.has_related_timesheets():
            self.fields["day"].help_text = nodate_change_warning
            self.fields["day"].widget.attrs["readonly"] = True

    day = SwissDateField(label=_("Date"))
    begin = SwissTimeField(label=_("Début"))
    address_canton = forms.ChoiceField(
        label=_("Canton"),
        widget=CHStateSelect,
        choices=STATE_CHOICES_WITH_DEFAULT,
        required=False,
    )
    apples = forms.CharField(
        label=_("Pommes"),
        required=False,
        widget=forms.TextInput(
            {
                "placeholder": _(
                    "Organisation de la livraison de pommes (quantité & logistique)"
                )
            }
        ),
    )
    helpers_time = SwissTimeField(
        label=_("Heure de rendez-vous des moniteur·trice·s"), required=False
    )
    superleader = UserAutoComplete(
        label=_("Moniteur·trice + / Photographe"),
        queryset=(get_user_model().objects.exclude(profile__status=USERSTATUS_DELETED)),
        url="user-AllPersons-ac",
        required=False,
    )
    bikes_phone = CHPhoneNumberField(label=_("N° de contact vélos"), required=False)

    def clean_day(self):
        day = self.cleaned_data["day"]

        if self.instance.day != day and self.instance.has_related_timesheets():
            raise forms.ValidationError(nodate_change_warning)

        if self.season.begin <= day <= self.season.end:
            return day
        raise forms.ValidationError(
            _("La session doit être dans le mois (entre {begin} et {end})").format(
                begin=date(self.season.begin, settings.DATE_FORMAT),
                end=date(self.season.end, settings.DATE_FORMAT),
            )
        )

    class Meta:
        model = Session
        labels = {
            "orga": _("Établissement"),
        }
        fields = [
            "orga",
            "day",
            "begin",
            "fallback_plan",
            "place",
            "address_street",
            "address_no",
            "address_zip",
            "address_city",
            "address_canton",
            "superleader",
            "apples",
            "helpers_time",
            "helpers_place",
            "bikes_concept",
            "bikes_phone",
            "comments",
            "visible",
        ]


class SessionDeleteForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        kwargs.pop("cantons", None)
        self.season = kwargs.pop("season", None)
        super().__init__(**kwargs)

    class Meta:
        model = Session
        fields = []



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2020 Sylvain Fankhauser <sylvain.fankhauser@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django import forms
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from apps.challenge.models import AnnualStateSetting, Season

from ..models import Invoice, InvoiceLine


class InvoiceFormMixin(forms.ModelForm):
    class Meta:
        model = Invoice
        fields = [
            "title",
            "status",
            "organization",
            "season",
        ]

    @transaction.atomic
    def save(self, commit=True):
        creating = not self.instance.pk
        season = self.cleaned_data["season"]
        if creating:
            ref_prefix = f"DV{season.year % 100}"
            try:
                largestref = (
                    Invoice.objects.filter(ref__startswith=ref_prefix)
                    .order_by("-ref")
                    .first()
                    .ref
                )
                startrefid = int(largestref[len(ref_prefix) :]) + 1
            except (AttributeError, ValueError):
                startrefid = 100
            for refid in range(startrefid, 999):
                ref = f"{ref_prefix}{refid:03d}"
                self.instance.ref = ref

                if not Invoice.objects.filter(ref=ref).exists():
                    # If unicity is garanteed, proceed
                    break
            else:
                # This should not happen.
                raise ValidationError(
                    _(
                        f"L’espace des factures établissements pour l’année {season.year} est épuisé avec l’identifiant {ref}."
                    )
                )
        invoice = super().save(commit=commit)

        try:
            settings = AnnualStateSetting.objects.get(
                canton=invoice.organization.address_canton, year=season.year
            )
        except AnnualStateSetting.DoesNotExist:
            settings = AnnualStateSetting()

        sessions = season.sessions.filter(orga=self.cleaned_data["organization"])

        if creating:
            for session in sessions:
                InvoiceLine.objects.create(
                    session=session,
                    historical_session=session.history.latest("history_date"),
                    invoice=invoice,
                    nb_participants=session.n_participants,
                    nb_bikes=session.n_bikes,
                    cost_bikes=(session.n_bikes * settings.cost_per_bike),
                    cost_participants=(
                        session.n_participants * settings.cost_per_participant
                    ),
                )
        elif not invoice.is_locked:
            # We're in update, but still in draft mode
            for session in sessions:
                InvoiceLine.objects.update_or_create(
                    session=session,
                    invoice=invoice,
                    defaults={
                        "nb_participants": session.n_participants,
                        "nb_bikes": session.n_bikes,
                        "cost_bikes": (session.n_bikes * settings.cost_per_bike),
                        "cost_participants": (
                            session.n_participants * settings.cost_per_participant
                        ),
                        "historical_session": session.history.latest("history_date"),
                    },
                )
            # Also delete superfluous lines
            InvoiceLine.objects.filter(invoice=invoice).exclude(
                session__in=sessions
            ).delete()
        return invoice


class InvoiceForm(InvoiceFormMixin):
    def __init__(self, organization, season, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for f in ["organization", "season"]:
            self.fields[f].disabled = True

        #  Reduce the number of convoluted queries. We know we want this one only.
        self.fields["season"].queryset = Season.objects.filter(pk=season.pk)

        # Circumvent weird bug due to the 'season' object being weird
        self.fields["season"].to_python = lambda x: x


class InvoiceFormQuick(InvoiceFormMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.fields:
            self.fields[field].widget = forms.HiddenInput()



# Generated by Django 1.11.4 on 2017-08-17 14:42
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0029_auto_20170814_1552"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="language",
            field=models.CharField(
                blank=True,
                choices=[
                    ("", "---------"),
                    ("fr", "French"),
                    ("de", "German"),
                    ("it", "Italian"),
                    ("en", "English"),
                ],
                max_length=7,
                verbose_name="Langue",
            ),
        ),
        migrations.AddField(
            model_name="userprofile",
            name="languages_challenges",
            field=models.CharField(
                blank=True,
                choices=[
                    ("fr", "French"),
                    ("de", "German"),
                    ("it", "Italian"),
                    ("en", "English"),
                ],
                max_length=11,
                verbose_name="Prêt à animer en",
            ),
        ),
    ]



# Generated by Django 1.11.4 on 2017-08-22 08:58
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0035_userprofile_formation_firstdate"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="tax_jurisdiction",
            field=models.CharField(
                blank=True,
                max_length=511,
                verbose_name="Lieu d'imposition (si pas en Suisse)",
            ),
        ),
    ]



from django import forms
from django.core.exceptions import ValidationError
from django.db.models import F
from django.utils.translation import gettext_lazy as _

from apps.challenge.models import Qualification, Season, Session
from apps.challenge.models.registration import Registration
from apps.common.forms import SwissDateField
from defivelo.templatetags.dv_filters import lettercounter


class OrganizationSelectionForm(forms.Form):
    def __init__(self, *args, coordinator, **kwargs):
        super().__init__(*args, **kwargs)
        organizations = coordinator.managed_organizations.all()
        self.fields["organization"] = forms.ModelChoiceField(
            widget=forms.Select,
            queryset=organizations,
            required=True,
            initial=organizations.first(),
            label=_("Établissement"),
        )


class RegistrationForm(forms.ModelForm):
    date = SwissDateField()

    def __init__(self, *args, **kwargs):
        coordinator = kwargs.pop("coordinator")
        # On GET request, organization is not yet chosen
        organization = kwargs.pop("organization", None)
        super().__init__(*args, **kwargs)
        self.instance.coordinator = coordinator
        self.instance.organization = organization
        self.fields["classes_amount"].widget.attrs["min"] = 1
        self.fields["classes_amount"].widget.attrs["max"] = 6

    def clean_date(self):
        """Check that a related Season(Month) exists"""
        data = self.cleaned_data["date"]
        if not Season.objects.filter(
            year=data.year,
            month_start__lte=data.month,
            n_months__gt=data.month - F("month_start"),
            cantons__contains=[self.instance.organization.address_canton],
        ).exists():
            raise ValidationError(
                _(
                    "Il n'est pas encore possible d'inscrire de session à cette date, "
                    "merci de contacter le ou la chargé·e de projet régional·e"
                )
            )

        return data

    class Meta:
        model = Registration
        fields = ("date", "day_time", "classes_amount")


class UniqueTimeRegistrationMixin:
    def clean(self):
        """Check that there's no two sessions at the same time."""
        if any(self.errors):
            # Don't bother validating the formset unless each form is valid on its own
            return

        times = []
        for form in self.forms:
            if self.can_delete and self._should_delete_form(form):
                continue
            time = "%s%s" % (
                form.cleaned_data.get("date").isoformat(),
                form.cleaned_data.get("day_time"),
            )
            if time in times:
                form.add_error(
                    "date",
                    _(
                        "Il est impossible de créer 2 sessions pour la même demi-journée pour le même établissement."
                    ),
                )
                form.add_error("day_time", "")

            times.append(time)


class BaseRegistrationFormSet(UniqueTimeRegistrationMixin, forms.BaseFormSet):
    def serialize(self):
        return [
            {
                "date": form["date"].isoformat(),
                "day_time": form["day_time"],
                "classes_amount": form["classes_amount"],
            }
            for form in self.cleaned_data
        ]


RegistrationFormSet = forms.formset_factory(
    RegistrationForm,
    formset=BaseRegistrationFormSet,
    extra=1,
    max_num=1,
    min_num=1,
    validate_min=True,
)


class RegistrationConfirmForm(forms.Form):
    is_terms_accepted = forms.BooleanField(
        label=_("J'accepte les conditions générales"),
        required=True,
        initial=False,
    )


class RegistrationValidationForm(forms.ModelForm):
    date = SwissDateField()
    is_validated = forms.BooleanField(
        initial=False,
        widget=forms.HiddenInput,
        required=True,
        error_messages={"required": _("Cette inscription n'a pas été validée.")},
    )

    def __init__(self, *args, **kwargs):
        organization = kwargs.pop("organization", None)
        super().__init__(*args, **kwargs)
        self.instance.organization = organization
        self.instance.coordinator = organization.coordinator

        self.fields["day_time"].widget.attrs["disabled"] = True
        self.fields["date"].widget.attrs["readonly"] = True
        self.fields["classes_amount"].widget.attrs["readonly"] = True

    def save(self, commit=True):
        instance = super().save(commit=commit)

        if self.cleaned_data["is_validated"]:
            session = Session.objects.create(
                day=instance.date,
                orga=instance.organization,
                begin=instance.day_time,
            )
            for i in range(instance.classes_amount):
                Qualification.objects.create(
                    session=session, name=_("Classe %s") % lettercounter(i + 1)
                )
            instance.archive()

    class Meta:
        model = Registration
        fields = ("id", "date", "day_time", "classes_amount")


class BaseRegistrationValidationFormSet(
    UniqueTimeRegistrationMixin, forms.BaseModelFormSet
):
    def __init__(self, *args, **kwargs):
        organization = kwargs.pop("organization")
        super().__init__(
            *args,
            queryset=Registration.objects.filter(
                organization=organization, is_archived=False
            ),
            form_kwargs={"organization": organization},
            **kwargs,
        )


RegistrationValidationFormSet = forms.modelformset_factory(
    Registration,
    form=RegistrationValidationForm,
    formset=BaseRegistrationValidationFormSet,
    extra=0,
    can_delete=True,
)



# Generated by Django 2.2.16 on 2020-11-03 09:34

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0070_add_months_columns"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="season",
            options={
                "ordering": ["year", "month_start", "cantons"],
                "verbose_name": "Saison",
                "verbose_name_plural": "Saisons",
            },
        ),
        migrations.RemoveField(
            model_name="historicalseason",
            name="season",
        ),
        migrations.RemoveField(
            model_name="season",
            name="season",
        ),
        migrations.AlterField(
            model_name="historicalseason",
            name="month_start",
            field=models.PositiveSmallIntegerField(
                choices=[
                    (1, "January"),
                    (2, "February"),
                    (3, "March"),
                    (4, "April"),
                    (5, "May"),
                    (6, "June"),
                    (7, "July"),
                    (8, "August"),
                    (9, "September"),
                    (10, "October"),
                    (11, "November"),
                    (12, "December"),
                ],
                default=1,
                verbose_name="Mois de début",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="season",
            name="month_start",
            field=models.PositiveSmallIntegerField(
                choices=[
                    (1, "January"),
                    (2, "February"),
                    (3, "March"),
                    (4, "April"),
                    (5, "May"),
                    (6, "June"),
                    (7, "July"),
                    (8, "August"),
                    (9, "September"),
                    (10, "October"),
                    (11, "November"),
                    (12, "December"),
                ],
                default=1,
                verbose_name="Mois de début",
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2.12 on 2020-05-05 14:20
from django.db import migrations
from django.utils.translation import gettext_lazy as _

USERSTATUS_DELETED = 99


def remove_email_from_deleted_users(apps, schema_editor):
    User = apps.get_model("auth", "User")
    for user in User.objects.filter(profile__status=USERSTATUS_DELETED):
        email_comment = _("Email avant suppression: %s") % user.email
        user.email = ""
        user.save()
        user.profile.comments = (
            f"{user.profile.comments}\n{email_comment}"
            if user.profile.comments
            else email_comment
        )
        user.profile.save()


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0054_timesheet_editor_perm"),
    ]

    operations = [
        migrations.RunPython(remove_email_from_deleted_users, migrations.RunPython.noop)
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015, 2016 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import urllib.parse
from re import search, sub

from django import template
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.template.defaultfilters import date as datefilter
from django.urls import reverse
from django.utils.dates import MONTHS
from django.utils.encoding import force_str
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _

import phonenumbers
from memoize import memoize
from phonenumbers import NumberParseException, PhoneNumberFormat
from rolepermissions.templatetags.permission_tags import can_template_tag

from apps.challenge import (
    AVAILABILITY_FIELDKEY,
    AVAILABILITY_FIELDKEY_HELPER_PREFIX,
    CHOICE_FIELDKEY,
    CHOSEN_AS_ACTOR,
    CHOSEN_AS_HELPER,
    CHOSEN_AS_LEADER,
    CHOSEN_AS_LEGACY,
    CHOSEN_AS_NOT,
    CHOSEN_AS_REPLACEMENT,
    CONFLICT_FIELDKEY,
    SEASON_WORKWISH_FIELDKEY,
    STAFF_FIELDKEY,
    STAFF_FIELDKEY_HELPER_PREFIX,
    SUPERLEADER_FIELDKEY,
)
from apps.common import (
    DV_SEASON_AUTUMN,
    DV_SEASON_CHOICES,
    DV_SEASON_LAST_SPRING_MONTH,
    DV_SEASON_SPRING,
    DV_STATE_CHOICES,
    DV_STATE_COLORS,
    STDGLYPHICON,
)
from apps.user import FORMATION_M1, FORMATION_M2, formation_short
from defivelo.roles import has_permission, user_cantons

register = template.Library()


@memoize()
def can_memoized(user, role):
    return can_template_tag(user, role)


# Override 'can' from rolepermissions to add memoization for performance reasons
@register.filter
def can(user, role):
    return can_memoized(user, role)


@register.simple_tag
def vcs_version():
    return settings.VCS_VERSION


@register.simple_tag
def vcs_commit():
    return settings.VCS_COMMIT


@register.filter
def setlang(request, newlang):
    """Replace language code in request.path with the new language code"""
    return sub("^/(%s)/" % request.LANGUAGE_CODE, "/%s/" % newlang, request.path)


def _format_phone_numer(
    tel: str | None, phone_format: int = PhoneNumberFormat.E164, region="CH"
) -> str:
    if not tel:
        return ""

    try:
        phone = phonenumbers.parse(tel, region)
        return phonenumbers.format_number(phone, phone_format)
    except NumberParseException:
        return tel


@register.filter
def tel_int(tel):
    if not tel:
        return ""

    # Already well formated
    if tel[0:3] == "+41" and len(tel) == 12:
        return tel

    return _format_phone_numer(tel)


@register.filter
def tel_link(tel):
    if not tel:
        return ""
    if tel[0:3] == "+41" and len(tel) == 12:
        # Convert to swiss local number
        tel_str = "0{} {} {} {}".format(tel[3:5], tel[5:8], tel[8:10], tel[10:12])
    else:
        tel_str = _format_phone_numer(tel, PhoneNumberFormat.NATIONAL)
    return mark_safe(
        '<a href="tel:{tel_int}">{tel}</a>'.format(
            tel_int=escape(tel_int(tel)), tel=escape(tel_str)
        )
    )


@register.filter
def profile_tag(user, is_restricted: bool = False):
    """
    Standard user display (currently fullname + small natel)
    """
    if not user:
        return ""
    usertag = "<span>"
    usertag += user.get_full_name()
    if user.profile.natel and not is_restricted:
        usertag += "<br /><small>{tel_link}</small>".format(
            tel_link=tel_link(user.profile.natel)
        )
    usertag += "</span>"

    return mark_safe(usertag)


@register.filter
def useravailsessions(form, user):
    """
    Output cells with form widget for all sessions concerning that user
    """
    if not form or not user:
        return ""
    output = ""
    for key in form.fields:
        if (
            key == SEASON_WORKWISH_FIELDKEY.format(hpk=user.pk)
            or AVAILABILITY_FIELDKEY_HELPER_PREFIX.format(hpk=user.pk) in key
        ):
            output += "<td>{field}</td>".format(
                field=form.fields[key].widget.render(
                    key, form.fields[key].initial, attrs={"id": key}
                )
            )
    return mark_safe(output)


@register.filter
def useravailsessions_readonly(
    struct,
    user,
    avail_forced_content=None,
    sesskey=None,
    onlyavail=False,
    planning=False,
):
    """
    Output cells with the state of the availability / choice for all sessions,
    for that user
    """
    if not struct or not user:
        return ""
    output = ""
    for key in struct:
        if AVAILABILITY_FIELDKEY_HELPER_PREFIX.format(hpk=user.pk) in key:
            # If there's a sessionkey specified, skip the ones not
            # corresponding
            if sesskey:
                if key != AVAILABILITY_FIELDKEY.format(hpk=user.pk, spk=sesskey):
                    continue
            availability = struct[key]
            avail_verb = ""  # Fulltext
            avail_label = ""  # Bootstrap glyphicon name
            avail_class = "active"  # Bootstrap array cell class
            avail_content = avail_forced_content
            conflict = False
            locked = False
            if not planning:
                if availability == "i":
                    # If needed
                    avail_verb = _("Si nécessaire")
                    avail_label = "ok-circle"
                    avail_class = "warning"
                elif availability == "n":
                    avail_verb = _("Non")
                    avail_label = "remove-sign"
                    avail_class = "danger"
                elif availability == "y":
                    avail_verb = _("Oui")
                    avail_label = "ok-sign"
                    avail_class = "success"

            thissesskey = int(search(r"-s(\d+)", key).group(1))

            slkey = SUPERLEADER_FIELDKEY.format(hpk=user.pk, spk=thissesskey)
            superleader = slkey in struct and struct[slkey]

            if availability in ["y", "i"]:
                choicekey = CHOICE_FIELDKEY.format(hpk=user.pk, spk=thissesskey)
                locked = choicekey in struct and struct[choicekey]

                conflictkey = CONFLICT_FIELDKEY.format(hpk=user.pk, spk=thissesskey)
                conflicts = struct[conflictkey] if conflictkey in struct else []
                if len(conflicts) > 0:
                    conflict = conflicts.pop()

                # Si le choix des moniteurs est connu, remplace le label et
                # la version verbeuse par l’état du choix
                if not sesskey:
                    staffkey = STAFF_FIELDKEY.format(hpk=user.pk, spk=thissesskey)
                    if staffkey in struct:
                        if struct[staffkey] == CHOSEN_AS_LEADER:
                            avail_verb = _("Moniteur·trice 2")
                            avail_content = formation_short(FORMATION_M2)
                        elif struct[staffkey] == CHOSEN_AS_HELPER:
                            avail_verb = _("Moniteur·trice 1")
                            avail_content = formation_short(FORMATION_M1)
                        elif struct[staffkey] == CHOSEN_AS_ACTOR:
                            avail_verb = _("Intervenant·e")
                            avail_label = "sunglasses"
                        elif struct[staffkey] == CHOSEN_AS_REPLACEMENT:
                            avail_verb = _("Moniteur·trice de secours")
                            avail_content = _("S")
                        elif struct[staffkey] == CHOSEN_AS_LEGACY:
                            avail_verb = _("Choisi")
                            avail_label = "check"
                        else:
                            avail_verb = _("Pas choisi")
                            if not planning:
                                avail_label = "unchecked"
                            else:
                                avail_label = "remove-sign"
                                avail_class = "danger"
            else:
                if availability == "n" and planning:
                    avail_label = "remove-sign"
                    avail_class = "danger"
                    avail_verb = _("Non")
                if onlyavail:
                    avail_content = " "
            final_avail_label = (
                avail_content
                if avail_content
                else (
                    STDGLYPHICON.format(
                        icon=avail_label, title=avail_verb if avail_verb else ""
                    )
                    if avail_label
                    else ""
                )
            )

            output += (
                '<td data-test="{key}" class="{avail_class}"{avail_verbose}><div class="dvflex">'
                "<!-- {key} -->{avail_label}{superleader}{conflict_warning}"
                "</div></td>"
            ).format(
                avail_class="info" if locked else avail_class,
                avail_verbose=' title="%s"' % avail_verb if avail_verb else "",
                avail_label=final_avail_label,
                superleader=(
                    ("&nbsp;/&nbsp;" if final_avail_label else "")
                    + '<span title="{mplus}">{mplus_short}</span>'.format(
                        mplus=_("Moniteur·trice + / Photographe"), mplus_short=_("M+")
                    )
                )
                if superleader and planning
                else "",
                conflict_warning=(
                    (
                        '<a class="text-danger" href="{season_url}#sess{sessionpk}">{glyph}</a>'.format(
                            season_url=reverse(
                                "season-availabilities",
                                kwargs={"pk": conflict.session.season.pk},
                            ),
                            sessionpk=conflict.session.pk,
                            glyph=STDGLYPHICON.format(
                                icon="alert", title=conflict.session
                            ),
                        )
                    )
                    if conflict and not planning
                    else ""
                ),
                key=key,
            )
    return mark_safe(output)


@register.filter
def userplanning_sessions_readonly(struct, user):
    """
    Output cells with the state of the availability / choice for all sessions,
    for that user
    """
    return useravailsessions_readonly(struct, user, planning=True)


@register.filter
def userstaffsessions(form, user):
    """
    Output cells with the form widgets for session choice
    for that user
    """
    if not form or not user:
        return ""
    output = ""
    for key in form.fields:
        if STAFF_FIELDKEY_HELPER_PREFIX.format(hpk=user.pk) in key:
            # Lookout for the session key in the form field key, and pipe
            # it to the useravailsessions_readonly function
            sesskey = int(search(r"-s(\d+)", key).group(1))
            output += useravailsessions_readonly(
                struct=form.initial,
                user=user,
                sesskey=sesskey,
                onlyavail=True,
                avail_forced_content=form.fields[key].widget.render(
                    key, form.fields[key].initial, attrs={"id": key}
                ),
            )
    return mark_safe(output)


@register.filter
def chosen_staff_for_season(struct, user):
    if not struct or not user:
        return ""
    accu_in_qualif = 0
    accu_in_sess = 0
    for key in struct:
        if AVAILABILITY_FIELDKEY_HELPER_PREFIX.format(hpk=user.pk) in key:
            if struct[key] in ["y", "i"]:
                thissesskey = int(search(r"-s(\d+)", key).group(1))
                choicekey = CHOICE_FIELDKEY.format(hpk=user.pk, spk=thissesskey)
                if choicekey in struct and struct[choicekey]:
                    accu_in_qualif += 1

                staffkey = STAFF_FIELDKEY.format(hpk=user.pk, spk=thissesskey)
                if staffkey in struct:
                    if struct[staffkey] != CHOSEN_AS_NOT:
                        accu_in_sess += 1
    #  if accu_in_qualif != accu_in_sess:
    #  return '%s/%s' % (accu_in_qualif, accu_in_sess)
    return accu_in_sess


@register.filter
def work_wish_for_season(struct, user):
    if not struct or not user:
        return ""
    workwishkey = SEASON_WORKWISH_FIELDKEY.format(hpk=user.pk)
    if workwishkey in struct:
        wish = int(struct[workwishkey])
        if wish > 0:
            return wish
    return ""


@register.filter
def weeknumber(date):
    if not date:
        return ""
    # This "solves" the weird week numbers in templates
    return date.strftime("%W")


@register.filter
def date_ch_short(date):
    if not date:
        return ""
    lang = get_language()
    if lang == "de":
        return datefilter(date, "j. N")
    else:
        return datefilter(date, "j N")


@register.filter
def cantons_abbr(cantons, abbr=True, long=True):
    return [
        force_str(c[1])
        if not abbr
        else mark_safe(
            '<abbr title="{title}">{abbr}</abbr>'.format(
                abbr=c[0],
                title=c[1],
            )
        )
        for c in DV_STATE_CHOICES
        if c[0] in cantons
    ]


@register.filter
def canton_abbr(canton, abbr=True, long=True):
    try:
        return cantons_abbr([canton], abbr, long)[0]
    except IndexError:
        return canton


@register.filter
def canton_abbr_short(canton, abbr=True):
    try:
        return cantons_abbr([canton], abbr=abbr, long=False)[0]
    except IndexError:
        return canton


@register.filter
def ifcanton_abbr(canton):
    try:
        return canton_abbr(canton)
    except IndexError:
        return canton


@register.filter
def season_verb(season_id):
    try:
        return [s[1] for s in DV_SEASON_CHOICES if s[0] == season_id][0]
    except IndexError:
        return ""


@register.simple_tag
def dv_season(day=None):
    """
    Structure (kwargs) for the current (or specified) DV season
    """
    if not day:
        day = datetime.datetime.today()
    return {
        "year": day.year,
        "dv_season": (
            DV_SEASON_SPRING
            if day.month <= DV_SEASON_LAST_SPRING_MONTH
            else DV_SEASON_AUTUMN
        ),
    }


@register.simple_tag
def dv_season_url(day=None):
    """
    URL of the current (or specified) DV season list view
    """
    return reverse("season-list", kwargs=dv_season(day))


@register.filter
def season_month_start(season_id):
    for s in DV_SEASON_CHOICES:
        if s[0] == season_id:
            try:
                return MONTHS[
                    1
                    + (
                        0
                        if season_id == DV_SEASON_SPRING
                        else DV_SEASON_LAST_SPRING_MONTH
                    )
                ]
            except IndexError:
                pass
    return ""


@register.filter
def season_month_end(season_id):
    for s in DV_SEASON_CHOICES:
        if s[0] == season_id:
            try:
                return MONTHS[
                    (
                        DV_SEASON_LAST_SPRING_MONTH
                        if season_id == DV_SEASON_SPRING
                        else 12
                    )
                ]
            except IndexError:
                pass
    return ""


@register.filter
def anyofusercantons(user, cantons):
    try:
        usercantons = user_cantons(user)
        return list(set(usercantons).intersection(set(cantons)))
    except PermissionDenied:
        return


@register.filter
def unprivileged_user_can_see(user, season):
    return season.unprivileged_user_can_see(user)


@register.filter
def inusercantons(user, canton):
    if has_permission(user, "cantons_all"):
        # Also True for `canton == ''`
        return True
    try:
        usercantons = user_cantons(user)
    except LookupError:
        return False
    else:
        return canton in usercantons


@register.filter
def lettercounter(count):
    icount = int(count)
    if 0 <= icount <= 26:
        return chr(64 + icount)
    return icount


@register.simple_tag
def canton_colors():
    return DV_STATE_COLORS


@register.filter
def remove_qs(url, param_name):
    """
    Remove all occurences of the querystring named `param_name` from the given URL.
    """
    parsed_url = urllib.parse.urlparse(url)
    parsed_qs = urllib.parse.parse_qsl(parsed_url.params)
    qs_without_param = [
        (name, value) for name, value in parsed_qs if name != param_name
    ]
    encoded_qs_without_param = urllib.parse.urlencode(qs_without_param)

    return urllib.parse.urlunparse(parsed_url._replace(query=encoded_qs_without_param))


@register.simple_tag
def add_qs(url, **kwargs):
    """
    Add all kwargs and their values as querystrings to the given URL.
    """
    parsed_url = urllib.parse.urlparse(url)
    new_qs = urllib.parse.parse_qsl(parsed_url.params) + [
        (name, value) for name, value in kwargs.items()
    ]
    new_qs_encoded = urllib.parse.urlencode(new_qs)

    return urllib.parse.urlunparse(parsed_url._replace(query=new_qs_encoded))


@register.filter
def get_timesheet_status_for_canton(mcv, timesheets_status):
    """
    mcv is the object
    timesheets_status is the canton's array of timesheet statuses
    """
    return timesheets_status.get(mcv.canton, None)



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import operator
from functools import reduce

from django.contrib.messages.views import SuccessMessageMixin
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView

from dal_select2.views import Select2QuerySetView
from django_filters import CharFilter, FilterSet, MultipleChoiceFilter
from django_filters.views import FilterView
from rolepermissions.checkers import has_role
from rolepermissions.mixins import HasPermissionsMixin

from apps.common import DV_STATE_CHOICES_WITH_DEFAULT
from apps.common.views import ExportMixin, PaginatorMixin
from defivelo.roles import has_permission, user_cantons
from defivelo.views import MenuView

from .export import OrganizationResource
from .forms import OrganizationForm
from .models import ORGASTATUS_ACTIVE, ORGASTATUS_CHOICES, Organization


class OrganizationFilterSet(FilterSet):
    def __init__(self, data=None, *args, **kwargs):
        any_filter_is_set = bool(set(self.base_filters) & set(data or {}))
        if not any_filter_is_set:
            data = {}
            for name, f in self.base_filters.items():
                initial = f.extra.get("initial")
                # filter param is either missing or empty, use initial as default
                if not data.get(name) and initial:
                    data[name] = initial

        super(OrganizationFilterSet, self).__init__(data, *args, **kwargs)

    def filter_wide(queryset, name, value):
        if value:
            allfields_filter = [
                Q(name__unaccent__icontains=value),
                Q(abbr__unaccent__icontains=value),
                Q(address_street__unaccent__icontains=value),
                Q(address_zip__contains=value),
                Q(address_city__unaccent__icontains=value),
            ]
            return queryset.filter(reduce(operator.or_, allfields_filter))
        return queryset

    address_canton = MultipleChoiceFilter(
        label=_("Cantons"),
        choices=DV_STATE_CHOICES_WITH_DEFAULT,
    )

    status = MultipleChoiceFilter(
        label=_("Statut"),
        choices=ORGASTATUS_CHOICES,
        initial=[
            ORGASTATUS_ACTIVE,
        ],
    )

    q = CharFilter(label=_("Recherche"), method=filter_wide)

    class Meta:
        model = Organization
        fields = [
            "q",
            "status",
            "address_canton",
        ]


class OrganizationDetailMixin(HasPermissionsMixin, MenuView):
    required_permission = "orga_show"
    model = Organization
    context_object_name = "organization"

    def get_context_data(self, **kwargs):
        context = super(OrganizationDetailMixin, self).get_context_data(**kwargs)
        # Add our menu_category context
        context["menu_category"] = "organization"
        return context


class OrganizationDetailView(OrganizationDetailMixin, DetailView):
    pass


class OrganizationsListView(OrganizationDetailMixin, PaginatorMixin, FilterView):
    filterset_class = OrganizationFilterSet
    context_object_name = "organizations"


class OrganizationListExport(ExportMixin, OrganizationsListView):
    export_class = OrganizationResource()
    export_filename = _("Établissements")


class OrganizationAutocomplete(OrganizationDetailMixin, Select2QuerySetView):
    required_permission = "orga_detail_all"

    def get_queryset(self):
        if has_permission(self.request.user, self.required_permission):
            qs = super(OrganizationAutocomplete, self).get_queryset()
            if self.q:
                qs = OrganizationFilterSet.filter_wide(qs, "", self.q)
            return qs
        else:
            raise PermissionDenied


class OrganizationCrudMixin(OrganizationDetailMixin):
    required_permission = "orga_crud"
    form_class = OrganizationForm

    def get_queryset(self):
        if self.request.user.managed_organizations.count():
            return self.request.user.managed_organizations
        qs = self.model.objects
        try:
            usercantons = user_cantons(self.request.user)
        except LookupError:
            raise PermissionDenied
        if usercantons:
            qs = qs.filter(address_canton__in=usercantons)
        return qs

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        try:
            kwargs["cantons"] = user_cantons(self.request.user)
        except LookupError:
            kwargs["cantons"] = [
                orga.address_canton
                for orga in self.request.user.managed_organizations.all()
            ]
        return kwargs

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        if has_role(self.request.user, "coordinator") and not has_role(
            self.request.user, "power_user"
        ):
            form.fields["status"].disabled = True
            form.fields["coordinator"].disabled = True
            del form.fields["comments"]
        if not has_permission(self.request.user, "orga_edit_address"):
            form.fields["address_canton"].disabled = True
        return form


class OrganizationUpdateView(OrganizationCrudMixin, SuccessMessageMixin, UpdateView):
    success_message = _("Établissement mis à jour")
    required_permission = "orga_edit"


class OrganizationCreateView(OrganizationCrudMixin, SuccessMessageMixin, CreateView):
    success_message = _("Établissement créé")



# Generated by Django 1.11.4 on 2017-08-22 13:16
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0038_multify_actor_for"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="userprofile",
            name="actor_for",
        ),
        migrations.RenameField(
            model_name="userprofile",
            old_name="actor_for_multi",
            new_name="actor_for",
        ),
    ]



# The MIT License (MIT)
#
# Copyright (c) 2015 Ha Pham
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from django.test import TestCase
from django.urls import reverse

from apps.article.models import Article
from defivelo.tests import utils


class ArticleModelTests(TestCase):
    def test_save(self):
        article = Article(title="Single One", summary="", body="Content")
        article.save()
        self.assertGreater(article.pk, 0)
        self.assertEqual(article.slug, "single-one")
        self.assertEqual(str(article), "Article: Single One")

    def test_duplicate(self):
        a0 = Article(title="Single One", summary="", body="Content")
        a0.save()
        a1 = Article(title="Single One", summary="", body="Content 2")
        a1.save()
        self.assertEqual(a1.slug, "single-one-1")

    def test_poweruser_only_can_crud_acrticles(self):
        article = Article(title="Single One", summary="", body="Content")
        article.save()

        # Any user
        self.client = utils.AuthClient()
        response = self.client.post(reverse("article-delete", args=[article.pk]))
        self.assertEqual(response.status_code, 403)
        self.assertEqual(Article.objects.count(), 1)

        # Power user
        self.client = utils.PowerUserAuthClient()
        response = self.client.post(reverse("article-delete", args=[article.pk]))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(Article.objects.count(), 0)



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015-2020 Didier Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django import forms
from django.contrib.auth import get_user_model
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from bootstrap3_datetime.widgets import DateTimePicker
from dal.forward import Const as dal_const
from dal_select2.widgets import ModelSelect2, Select2Multiple

from apps.common import (
    DV_SEASON_STATE_OPEN,
    DV_SEASON_STATE_RUNNING,
    DV_SEASON_STATES,
    DV_STATE_CHOICES_WITH_ABBR,
)
from apps.common.forms import SelectWithDisabledValues
from apps.user import FORMATION_KEYS, FORMATION_M2
from apps.user.models import USERSTATUS_DELETED

from ...orga.models import Organization
from .. import (
    AVAILABILITY_FIELDKEY,
    CHOICE_CHOICES,
    CHOSEN_AS_ACTOR,
    CHOSEN_AS_HELPER,
    CHOSEN_AS_LEADER,
    CHOSEN_AS_NOT,
    CHOSEN_AS_REPLACEMENT,
    SEASON_WORKWISH_FIELDKEY,
    STAFF_FIELDKEY,
)
from ..fields import BSAvailabilityRadioSelect, BSChoiceRadioSelect, LeaderChoiceField
from ..models import Season
from ..models.availability import HelperSessionAvailability


class SeasonForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        cantons = kwargs.pop("cantons", None)
        kwargs.pop("season", None)
        super().__init__(**kwargs)
        self.fields["state"].widget = SelectWithDisabledValues(
            choices=self.fields["state"].choices,
            disabled_values=[DV_SEASON_STATE_RUNNING, DV_SEASON_STATE_OPEN],
        )

        if cantons:
            # Only permit edition within the allowed cantons
            choices = self.fields["cantons"].choices
            choices = ((k, v) for (k, v) in choices if k in cantons)
            self.fields["cantons"].choices = choices

    year = forms.IntegerField(
        label=_("Année"),
        widget=DateTimePicker({"placeholder": "YYYY"}, options={"format": "YYYY"}),
    )
    leader = LeaderChoiceField(
        label=_("Chargé·e de projet"),
        queryset=(
            get_user_model()
            .objects.exclude(profile__status=USERSTATUS_DELETED)
            .filter(managedstates__isnull=False)
            .distinct()
        ),
        required=True,
    )

    class Meta:
        model = Season
        fields = ["year", "month_start", "n_months", "cantons", "state", "leader"]


class SeasonToSpecificStateForm(forms.ModelForm):
    """
    A form with hidden fields and a tickbox that will _only_ update a season to a specific season state
    """

    sendemail = forms.BooleanField(
        label=_("Envoyer le courriel suivant"), initial=True, required=False
    )
    customtext = forms.CharField(
        label=_("Précisions"),
        help_text=_(
            "Attention: Le courriel est envoyé à chaque destinataire dans sa langue, mais ce texte est envoyé tel quel, il ne sera pas traduit."
        ),
        initial="",
        widget=forms.Textarea,
        required=False,
    )

    def __init__(
        self,
        tostate: int,
        *args,
        **kwargs,
    ):
        """
        Takes a tostate argument, an int of desired target state
        """
        kwargs.pop("cantons", None)
        kwargs.pop("season", None)
        super().__init__(*args, **kwargs)
        # Only allow one target state
        state_choices = self.fields["state"].choices
        assert tostate in [k for (k, v) in DV_SEASON_STATES]
        self.fields["state"].choices = (
            (k, v) for (k, v) in state_choices if k == tostate
        )
        self.fields["state"].widget = forms.HiddenInput()

    class Meta:
        model = Season
        fields = ["state", "sendemail"]


class SeasonAvailabilityReminderForm(forms.ModelForm):
    """
    A form to trigger availability reminder emails for a season.
    """

    sendemail = forms.BooleanField(
        label=_("Envoyer le courriel suivant"), initial=True, required=False
    )

    def __init__(self, *args, **kwargs):
        kwargs.pop("cantons", None)
        kwargs.pop("season", None)
        super().__init__(*args, **kwargs)

    class Meta:
        model = Season
        fields: list[str] = []


class SeasonNewHelperAvailabilityForm(forms.Form):
    def __init__(self, *args, **kwargs):
        cantons = kwargs.pop("cantons", [])
        super(SeasonNewHelperAvailabilityForm, self).__init__(*args, **kwargs)
        self.fields["helper"] = forms.ModelChoiceField(
            label=_("Disponibilités pour :"),
            queryset=get_user_model()
            .objects.filter(
                Q(profile__formation__in=FORMATION_KEYS)
                | Q(profile__actor_for__isnull=False)
            )
            .exclude(profile__status=USERSTATUS_DELETED)
            .distinct(),
            widget=ModelSelect2(
                url="user-PersonsRelevantForSessions-ac",
                forward=[
                    dal_const(cantons, "cantons"),
                ],
            ),
        )


class SeasonAvailabilityForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.season = kwargs.pop("instance")
        self.potential_helpers = kwargs.pop("potential_helpers")
        kwargs.pop("season", None)
        kwargs.pop("cantons", None)
        super(SeasonAvailabilityForm, self).__init__(*args, **kwargs)

        if self.potential_helpers:
            for helper_category, helpers in self.potential_helpers:
                for helper in helpers:
                    workwishkey = SEASON_WORKWISH_FIELDKEY.format(hpk=helper.pk)
                    try:
                        fieldinit = self.initial[workwishkey]
                    except KeyError:
                        fieldinit = 0
                    self.fields[workwishkey] = forms.IntegerField(
                        required=False,
                        initial=fieldinit,
                        min_value=0,
                    )

                    for session in self.season.sessions_with_qualifs:
                        availkey = AVAILABILITY_FIELDKEY.format(
                            hpk=helper.pk, spk=session.pk
                        )
                        staffkey = STAFF_FIELDKEY.format(hpk=helper.pk, spk=session.pk)
                        try:
                            fieldinit = self.initial[availkey]
                        except KeyError:
                            fieldinit = ""
                        try:
                            forbid_absence = self.initial[staffkey]
                        except KeyError:
                            forbid_absence = False
                        # Trick to pass the 'chosen' information through
                        self.fields[availkey] = forms.ChoiceField(
                            choices=HelperSessionAvailability.AVAILABILITY_CHOICES,
                            widget=BSAvailabilityRadioSelect(
                                forbid_absence=forbid_absence
                            ),
                            required=False,
                            initial=fieldinit,
                        )

    def save(self):
        pass


class SeasonStaffFilterForm(forms.Form):
    cantons = forms.MultipleChoiceField(
        label=_("Cantons"),
        choices=DV_STATE_CHOICES_WITH_ABBR,
        widget=Select2Multiple(attrs={"data-placeholder": _("Tous les cantons")}),
        required=False,
    )

    organisations = forms.MultipleChoiceField(
        label=_("Établissements"),
        choices=[],
        widget=Select2Multiple(
            attrs={"data-placeholder": _("Tous les établissements")}
        ),
        required=False,
    )

    def __init__(self, organisations: list[Organization], *args, **kwargs):
        super().__init__(*args, **kwargs)
        if organisations:
            self.fields["organisations"].choices = [
                (orga.pk, orga.abbr_or_name)
                for orga in sorted(organisations, key=lambda o: o.abbr_or_name)
            ]
        else:
            del self.fields["organisations"]


class SeasonStaffChoiceForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.season = kwargs.pop("instance")
        self.available_helpers = kwargs.pop("available_helpers")
        kwargs.pop("season", None)
        kwargs.pop("cantons", None)
        super(SeasonStaffChoiceForm, self).__init__(*args, **kwargs)

        if self.available_helpers:
            for helper_category, helpers in self.available_helpers:
                for helper in helpers:
                    for session in self.season.sessions_with_qualifs:
                        AVAILABILITY_FIELDKEY.format(hpk=helper.pk, spk=session.pk)
                        staffkey = STAFF_FIELDKEY.format(hpk=helper.pk, spk=session.pk)
                        try:
                            fieldinit = self.initial[staffkey]
                        except KeyError:
                            fieldinit = CHOSEN_AS_NOT
                        available_choices = [CHOSEN_AS_NOT]
                        if helper.profile.actor:
                            available_choices.append(CHOSEN_AS_ACTOR)
                        if helper.profile.formation:
                            available_choices.append(CHOSEN_AS_HELPER)
                            available_choices.append(CHOSEN_AS_REPLACEMENT)
                        if helper.profile.formation == FORMATION_M2:
                            available_choices.append(CHOSEN_AS_LEADER)
                        self.fields[staffkey] = forms.ChoiceField(
                            choices=[
                                c
                                for c in list(CHOICE_CHOICES)
                                if c[0] in available_choices
                            ],
                            widget=BSChoiceRadioSelect(
                                attrs={"horizontal": True, "class": "btn-group-xs"},
                                user_assignment=session.user_assignment(helper),
                            ),
                            required=False,
                            initial=fieldinit,
                        )

    def save(self):
        pass



# Generated by Django 3.2.13 on 2022-04-19 08:25
from django.db import migrations, models
from django.db.models import F, Value
from django.db.models.functions import Concat

import apps.common.fields


def forward(apps, schema_editor):
    """
    Reformat data for the new FieldType
    """
    Season = apps.get_model("challenge", "Season")
    Season.objects.update(cantons=Concat(Value("{"), F("cantons"), Value("}")))

    HSeason = apps.get_model("challenge", "HistoricalSeason")
    HSeason.objects.update(cantons=Concat(Value("{"), F("cantons"), Value("}")))


class Migration(migrations.Migration):
    dependencies = [
        ("challenge", "0079_add_zoug"),
    ]

    operations = [
        migrations.RunPython(forward),
        migrations.AlterField(
            model_name="historicalseason",
            name="cantons",
            field=apps.common.fields.ChoiceArrayField(
                base_field=models.CharField(
                    choices=[
                        ("AG", "Aargau"),
                        ("AR", "Appenzell Ausserrhoden"),
                        ("BE", "Berne"),
                        ("BL", "Basel-Land"),
                        ("BS", "Basel-Stadt"),
                        ("FR", "Fribourg"),
                        ("GE", "Geneva"),
                        ("GR", "Graubuenden"),
                        ("JU", "Jura"),
                        ("LU", "Lucerne"),
                        ("NE", "Neuchatel"),
                        ("SG", "St. Gallen"),
                        ("SO", "Solothurn"),
                        ("SZ", "Schwyz"),
                        ("VD", "Vaud"),
                        ("VS", "Valais"),
                        ("ZG", "Zug"),
                        ("ZH", "Zurich"),
                    ],
                    max_length=2,
                ),
                default=list,
                size=None,
                verbose_name="Cantons",
            ),
        ),
        migrations.AlterField(
            model_name="season",
            name="cantons",
            field=apps.common.fields.ChoiceArrayField(
                base_field=models.CharField(
                    choices=[
                        ("AG", "Aargau"),
                        ("AR", "Appenzell Ausserrhoden"),
                        ("BE", "Berne"),
                        ("BL", "Basel-Land"),
                        ("BS", "Basel-Stadt"),
                        ("FR", "Fribourg"),
                        ("GE", "Geneva"),
                        ("GR", "Graubuenden"),
                        ("JU", "Jura"),
                        ("LU", "Lucerne"),
                        ("NE", "Neuchatel"),
                        ("SG", "St. Gallen"),
                        ("SO", "Solothurn"),
                        ("SZ", "Schwyz"),
                        ("VD", "Vaud"),
                        ("VS", "Valais"),
                        ("ZG", "Zug"),
                        ("ZH", "Zurich"),
                    ],
                    max_length=2,
                ),
                default=list,
                size=None,
                verbose_name="Cantons",
            ),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015, 2020 Didier Raboud <me+defivelo@odyx.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from django.conf import settings
from django.contrib import admin
from django.templatetags.static import static
from django.utils.safestring import mark_safe

from parler.admin import TranslatableAdmin
from simple_history.admin import SimpleHistoryAdmin

from .models import (
    AnnualStateSetting,
    HelperSessionAvailability,
    Qualification,
    QualificationActivity,
    Season,
    Session,
)
from .models.registration import Registration


class QualificationActivityAdmin(TranslatableAdmin):
    list_display = ["category", "name", "missing_languages"]
    list_display_links = ("name",)
    search_fields = ("translations__name", "pk")

    def missing_languages(self, obj):
        missing_languages = list(
            set([lang[0] for lang in settings.LANGUAGES])
            - set(obj.get_available_languages())
        )
        if missing_languages:
            return " | ".join(missing_languages)
        else:
            return mark_safe('<img src="%s" />' % static("admin/img/icon-yes.svg"))


class SeasonAdmin(SimpleHistoryAdmin):
    pass


class SessionAdmin(SimpleHistoryAdmin):
    list_display = (
        "day",
        "begin",
        "canton",
        "orga",
    )
    list_filter = (
        "orga",
        "day",
        "begin",
    )
    search_fields = ("orga",)

    def canton(self, obj):
        return obj.orga.address_canton


class QualificationAdmin(SimpleHistoryAdmin):
    pass


class HelperSessionAvailabilityAdmin(SimpleHistoryAdmin):
    pass


class AnnualStateSettingAdmin(SimpleHistoryAdmin):
    pass


admin.site.register(QualificationActivity, QualificationActivityAdmin)
admin.site.register(Season, SeasonAdmin)
admin.site.register(Session, SessionAdmin)
admin.site.register(Qualification, QualificationAdmin)
admin.site.register(HelperSessionAvailability, HelperSessionAvailabilityAdmin)
admin.site.register(AnnualStateSetting, AnnualStateSettingAdmin)
admin.site.register(Registration)



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2015,2020 Didier Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.forms import ModelChoiceField, Select, TimeField
from django.template.loader import render_to_string

from bootstrap3_datetime.widgets import DateTimePicker
from dal_select2.widgets import ModelSelect2
from django_countries import countries
from django_countries.fields import LazyTypedChoiceField
from django_countries.widgets import CountrySelectWidget
from localflavor.generic.forms import DEFAULT_DATE_INPUT_FORMATS, DateField
from phonenumber_field.formfields import PhoneNumberField
from phonenumber_field.widgets import RegionalPhoneNumberWidget

SWISS_DATE_INPUT_FORMAT = "%d.%m.%Y"
SWISS_DATE_DISPLAY_FORMAT = "DD.MM.YYYY"

SWISS_DATE_INPUT_FORMATS = DEFAULT_DATE_INPUT_FORMATS + (SWISS_DATE_INPUT_FORMAT,)


class SwissDateField(DateField):
    """
    A date input field which uses the bootstrap widget for pickin a Date
    """

    def __init__(self, input_formats=None, *args, **kwargs):
        input_formats = input_formats or SWISS_DATE_INPUT_FORMATS
        super(DateField, self).__init__(
            input_formats=input_formats,
            widget=DateTimePicker(
                {"placeholder": SWISS_DATE_DISPLAY_FORMAT},
                options={"format": SWISS_DATE_DISPLAY_FORMAT},
            ),
            *args,
            **kwargs,
        )


class SwissTimeInput(DateTimePicker):
    format = "HH:mm"

    def __init__(self, attrs={}, *args, **kwargs):
        attrs["placeholder"] = "HH:mm"
        super().__init__(
            attrs,
            icon_attrs={"class": "glyphicon glyphicon-time"},
            options={"format": "HH:mm", "stepping": 5},
            *args,
            **kwargs,
        )


class SwissTimeField(TimeField):
    """
    A date input field which uses the bootstrap widget for pickin a Date
    """

    def __init__(self, *args, **kwargs):
        super(TimeField, self).__init__(
            widget=SwissTimeInput,
            *args,
            **kwargs,
        )


class CHPhoneNumberField(PhoneNumberField):
    """
    A PhoneNumberField that uses the national fallback widget
    """

    def __init__(self, *args, **kwargs):
        super(CHPhoneNumberField, self).__init__(
            widget=RegionalPhoneNumberWidget, *args, **kwargs
        )


class BS3CountriesField(LazyTypedChoiceField):
    """
    A Bootstrap3 Countries selection Field
    """

    def __init__(self, *args, **kwargs):
        super(BS3CountriesField, self).__init__(
            widget=CountrySelectWidget(
                layout=render_to_string("country_select_widget.html")
            ),
            choices=countries,
            initial="CH",
            *args,
            **kwargs,
        )


class UserAutoComplete(ModelChoiceField):
    """
    A User input field which uses the Autocmplete URL and has the good
    default widget
    """

    def __init__(self, *args, **kwargs):
        url = kwargs.pop("url", False)
        super(UserAutoComplete, self).__init__(
            widget=ModelSelect2(url=url), *args, **kwargs
        )

    def label_from_instance(self, obj):
        return obj.get_full_name()


class SelectWithDisabledValues(Select):
    """
    Select widget allowing disabled values
    """

    def __init__(self, *args, **kwargs):
        self.disabled_values = kwargs.pop("disabled_values", [])
        super().__init__(**kwargs)

    def create_option(self, name, value, label, selected, *args, **kwargs):
        """
        Mark disabled_values as such (unless selected)
        """
        option = super().create_option(name, value, label, selected, *args, **kwargs)
        if option.get("value") is None:
            option["attrs"]["disabled"] = "disabled"
        if not selected and option.get("value") in self.disabled_values:
            option["attrs"]["disabled"] = "disabled"
        return option



from django.db import migrations


def ws_to_vs(apps, schema_editor):
    Address = apps.get_model("common", "Address")
    Address.objects.filter(address_canton="WS").update(address_canton="VS")


class Migration(migrations.Migration):
    dependencies = [
        ("common", "0007_use_u2019_quote"),
    ]

    operations = [
        migrations.RunPython(ws_to_vs, migrations.RunPython.noop),
    ]



from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("user", "0011_auto_20151009_1532"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="birthdate",
            field=models.DateField(verbose_name="Date", null=True, blank=True),
        ),
    ]



# Generated by Django 2.2.12 on 2020-05-18 12:34

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "salary",
            "0009_monthlycantonalvalidationurl_monthlycantonalvalidationurltranslation",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="monthlycantonalvalidation",
            name="validated_urls",
            field=models.ManyToManyField(to="salary.MonthlyCantonalValidationUrl"),
        ),
    ]



# defivelo-intranet -- Outil métier pour la gestion du Défi Vélo
# Copyright (C) 2020 Didier Raboud <didier.raboud@liip.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .invoices import InvoiceForm, InvoiceFormQuick
from .qualification import (
    QualificationDeleteForm,
    QualificationForm,
    QualificationFormQuick,
)
from .season import (
    SeasonAvailabilityForm,
    SeasonAvailabilityReminderForm,
    SeasonForm,
    SeasonNewHelperAvailabilityForm,
    SeasonStaffChoiceForm,
    SeasonToSpecificStateForm,
)
from .session import SessionForm
from .settings import AnnualStateSettingForm

__all__ = [
    "AnnualStateSettingForm",
    "InvoiceForm",
    "InvoiceFormQuick",
    "QualificationDeleteForm",
    "QualificationForm",
    "QualificationFormQuick",
    "SeasonAvailabilityForm",
    "SeasonAvailabilityReminderForm",
    "SeasonForm",
    "SeasonNewHelperAvailabilityForm",
    "SeasonStaffChoiceForm",
    "SeasonToSpecificStateForm",
    "SessionForm",
]
